/*==========================================================================
 * Copyright (c) 20014 Carnegie Mellon University.  All Rights Reserved.
 *
 * Use of the Lemur Toolkit for Language Modeling and Information Retrieval
 * is subject to the terms of the software license set forth in the LICENSE
 * file included with this software (and below), and also available at
 * http://www.lemurproject.org/license.html
 *
 *==========================================================================
*/

#include "FeatureGenerator.h"
#include "plsa.h"
#include "Utils.h"
#include <algorithm>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <ctime>
#include <cstdlib>
#include <string>
#include <queue>

#include <omp.h>

FeatureGenerator::FeatureGenerator(std::vector<std::string>& idx_paths, int qid, std::string &q_str, std::string &res_path, std::string &phrase_file)
{
    queryid = qid;
    for(int i = 0; i<idx_paths.size();i++)
    {
        env.addIndex(idx_paths[i]);
    }

    svd = NULL;
    stem_document_matrix = NULL;
    query_str = q_str;

    // Read intial retrieval results
    std::vector<std::string> docnames;
    std::ifstream res_in(res_path.c_str(), std::ifstream::in);
    std::string line;

    while(getline(res_in, line))
    {
        std::istringstream iss(line);
        float qqid;
        std::string temp;
        std::string docname;
        int rank;
        float score;
        iss>>qqid>>temp>>docname>>rank>>score;
        docnames.push_back(docname);
        res_scores.push_back(score);
    }

    res_in.close();
    res_docids = env.documentIDsFromMetadata("docno", docnames);    

    for(int i=0;i<res_docids.size();i++)
    {
        res_docids_set.insert(res_docids[i]);
    }

    
    // Read phrase file
    std::ifstream phrase_in(phrase_file.data(), std::ifstream::in);
    while(getline(phrase_in, line))
    {
        std::istringstream iss(line);
        std::string temp;
        iss>>temp;
        if(temp[0]=='#')
        {
            // First experiments, no phrases
            output_phrases.push_back(temp);
        }
        else
            output_terms.push_back(temp);
    }

    phrase_in.close();

}

FeatureGenerator::FeatureGenerator(std::vector<std::string>& idx_paths, int qid, std::string &q_str, std::string &res_path) 
{
    queryid = qid;
    for(int i = 0; i<idx_paths.size();i++)
    {
        env.addIndex(idx_paths[i]);
    }

    svd = NULL;
    stem_document_matrix = NULL;
    query_str = q_str;

    // Read intial retrieval results
    std::vector<std::string> docnames;
    std::ifstream res_in(res_path.c_str(), std::ifstream::in);
    std::string line;
    while(getline(res_in, line))
    {
        std::istringstream iss(line);
        int qqid;
        std::string temp;
        std::string docname;
        int rank;
        float score;
        iss>>qqid>>temp>>docname>>rank>>score;
        if(qqid!=qid)
        {
            // Has completed reading res for qid
            if(docnames.size()!=0)
            {
                break;
            }
            else
            {
                continue;
            }
        }
        docnames.push_back(docname);
        res_scores.push_back(score);
    }

    res_in.close();

    res_docids = env.documentIDsFromMetadata("docno", docnames);    

    for(int i=0;i<res_docids.size();i++)
    {
        res_docids_set.insert(res_docids[i]);
    }

}

FeatureGenerator::FeatureGenerator(std::string& idx_path, int qid, std::string &q_str, std::string &res_path) 
{
    queryid = qid;
    env.addIndex(idx_path);
    svd = NULL;
    stem_document_matrix = NULL;
    query_str = q_str;

    // Read intial retrieval results
    std::vector<std::string> docnames;
    std::ifstream res_in(res_path.c_str(), std::ifstream::in);
    std::string line;
    while(getline(res_in, line))
    {
        std::istringstream iss(line);
        int qqid;
        std::string temp;
        std::string docname;
        int rank;
        float score;
        iss>>qqid>>temp>>docname>>rank>>score;
        if(qqid!=qid)
        {
            // Has completed reading res for qid
            if(docnames.size()!=0)
            {
                break;
            }
            else
            {
                continue;
            }
        }
        docnames.push_back(docname);
        res_scores.push_back(score);
    }

    res_in.close();

    res_docids = env.documentIDsFromMetadata("docno", docnames);  

    for(int i=0;i<res_docids.size();i++)
    {
        res_docids_set.insert(res_docids[i]);
    }

}


FeatureGenerator::FeatureGenerator(std::string& idx_path, int qid, std::string &q_str, std::string &pos_path, std::string &res_path) 
{
    queryid = qid;
    env.addIndex(idx_path);
    svd = NULL;
    stem_document_matrix = NULL;
    query_str = q_str;

    // Read intial retrieval results
    std::vector<std::string> docnames;
    std::ifstream res_in(res_path.c_str(), std::ifstream::in);
    std::string line;
    while(getline(res_in, line))
    {
        std::istringstream iss(line);
        int qqid;
        std::string temp;
        std::string docname;
        int rank;
        float score;
        iss>>qqid>>temp>>docname>>rank>>score;
        if(qqid!=qid)
        {
            // Has completed reading res for qid
            if(docnames.size()!=0)
            {
                break;
            }
            else
            {
                continue;
            }
        }
        docnames.push_back(docname);
        res_scores.push_back(score);
    }

    res_in.close();

    res_docids = env.documentIDsFromMetadata("docno", docnames);    

    for(int i=0;i<res_docids.size();i++)
    {
        res_docids_set.insert(res_docids[i]);
    }

    // Read POS results
    std::ifstream pos_in(pos_path.data(), std::ifstream::in);
    getline(pos_in, line);
    std::vector<std::string> parts = split(line, ' ');
    std::vector<bool> output_term_flags;
    std::vector<std::string> output_term_temp;
    for(int i=0;i<parts.size();i++)
    {
        // e.g.: What/WP sites/NNS give/VBP me/PRP information/NN about/IN public/JJ radio/NN and/CC TV/NN stations/NNS
        std::vector<std::string> term_parts = split(parts[i], '/');
        output_term_temp.push_back(term_parts[0]);
        output_term_flags.push_back(false);
        if(term_parts[1].compare("NNS")==0||term_parts[1].compare("NNP")==0||term_parts[1].compare("NNPS")==0)
            output_term_tags.push_back("NN");
        else
            output_term_tags.push_back(term_parts[1]);
    }
    pos_in.close();
    
    // Get phrases
    int phrase_size = 0;
    for(int i=0;i<output_term_tags.size();i++)
    {
        if(output_term_tags[i].compare("NN")==0)
            phrase_size ++;

        if(phrase_size >= 2 && (output_term_tags[i].compare("NN")!=0 || i==output_term_tags.size() -1))
        {
                int start, end;
                // [start, end)
                if(output_term_tags[i].compare("NN")!=0)
                {
                    start = i - phrase_size;
                    end = i;
                }

                if(i == output_term_tags.size() -1 && output_term_tags[i].compare("NN")==0)
                {
                    start = i - phrase_size + 1;
                    end = i + 1;
                }

                std::string stemmed_phrase("");
                std::string phrase("#1( ");
                
                for(int j=start;j<end;j++)
                {
                    stemmed_phrase += (stem_it(output_term_temp[j]) + " ");
                    phrase += (output_term_temp[j] + " ");

                    // Mark these terms as part of phrases
                    output_term_flags[j] = true;
                }
                
                phrase += ")";
                
                if(output_phrase_set.find(stemmed_phrase)==output_phrase_set.end()) // New phrase
                {
                    output_phrase_set.insert(stemmed_phrase);
                    output_phrases.push_back(phrase);
                }
                //   }
            
            phrase_size = 0;
        }

        if(output_term_tags[i].compare("NN")!=0)
            phrase_size = 0;
    }

    // Do not output features for terms in phrases
    // But this works worse than compute features for terms in phrases
    for(int i=0;i<output_term_temp.size();i++)
        //if(output_term_flags[i]==false)
        output_terms.push_back(output_term_temp[i]);
    
}

FeatureGenerator::FeatureGenerator(std::string& idx_path, int qid, std::string &q_str, std::string &query_part_file, std::string &res_path, bool flag_for_reading_query_part_file_format) 
{
    queryid = qid;
    env.addIndex(idx_path);
    svd = NULL;
    stem_document_matrix = NULL;
    query_str = q_str;

    // Read intial retrieval results
    std::vector<std::string> docnames;
    std::ifstream res_in(res_path.c_str(), std::ifstream::in);
    std::string line;
    while(getline(res_in, line))
    {
        std::istringstream iss(line);
        int qqid;
        std::string temp;
        std::string docname;
        int rank;
        float score;
        iss>>qqid>>temp>>docname>>rank>>score;
        if(qqid!=qid)
        {
            // Has completed reading res for qid
            if(docnames.size()!=0)
            {
                break;
            }
            else
            {
                continue;
            }
        }
        docnames.push_back(docname);
        res_scores.push_back(score);
    }

    res_in.close();

    res_docids = env.documentIDsFromMetadata("docno", docnames);    

    for(int i=0;i<res_docids.size();i++)
    {
        res_docids_set.insert(res_docids[i]);
    }

    // Read query part file
    std::ifstream query_part_in(query_part_file.data(), std::ifstream::in);
    while(getline(query_part_in, line))
    {
        std::string expr = line;
        if(expr[0]=='#')
        {
            std::string stemmed_expr = stem_expr(expr);
            if(output_phrase_set.find(stemmed_expr) == output_phrase_set.end()) // New expr
            {
                output_phrase_set.insert(stemmed_expr);
                output_phrases_unique.push_back(expr);
            }
            output_phrases.push_back(expr);
        }
        else
        {
            output_terms.push_back(expr);
            output_stem_set.insert(stem_it(expr));
        }
    }

    query_part_in.close();
}

FeatureGenerator::~FeatureGenerator()
{
    env.close();
    
    if(stem_document_matrix!=NULL)
    {
        delete stem_document_matrix;
    }
    
    if(svd!=NULL)
    {
        delete svd;
    }
}

void FeatureGenerator::setNumOfTopDocuments(int ntop)
{
    n_top = ntop;
}

std::string FeatureGenerator::stem_expr(std::string &expr) // #1(a b) or #uw8(a b)
{
    int left_p = expr.find('(');
    int right_p = expr.find(')');

    std::string raw_expr = expr.substr(left_p + 1, right_p - left_p - 1);
    std::vector<std::string> terms = split(raw_expr, ' ');

    std::string stemmed_expr = expr.substr(0, left_p); // #1 or #uw8
    stemmed_expr += '(';
    for(int i=0;i<terms.size();i++)
        if(terms[i].compare("")!=0)
        {
            stemmed_expr += stem_it(terms[i]);
            if(i!=terms.size()-1)
                stemmed_expr += ' ';
        }
    stemmed_expr += ')';

    return stemmed_expr;
}

std::string FeatureGenerator::stem_it(std::string &term)
{
    std::string stem = env.stemTerm(term);
    stem_term_map[stem] = term;
    return stem;
}

std::string tostr(int a)
{
    std::ostringstream oss;
    oss<<a;
    return oss.str();
}


std::vector<int> FeatureGenerator::computeExpressionTfs(std::vector<std::string> &expressions, indri::api::DocumentVector* dv)
{
    std::vector<int> tfs(expressions.size(), 0);
    
    for(int i=0;i<dv->positions().size();i++)
    {
        int pos_stem = dv->positions()[i];
        std::string stem = dv->stems()[pos_stem];

        //#1
        if(i<dv->positions().size()-1)
        {
            int pos_stem_after = dv->positions()[i+1];
            std::string stem_after = dv->stems()[pos_stem_after];
            
            std::string phrase("#1(");
            phrase += (stem + ' ' + stem_after + ')');
            for(int k = 0;k<expressions.size();k++)
            {
                std::string expression = expressions[k];
                if(phrase.compare(stem_expr(expression)) == 0)
                {
                    tfs[k]++;
                }
            }
        }

        //#uw8
        //range: [j_start, j_end]
        int j_start = i - 8 < 0 ? 0: i - 8;
        int j_end = i + 8 > dv->positions().size() -1 ? dv->positions().size() -1: i + 8;
        
        for(int j = j_start; j<= j_end && j!=i ; j++)
        {
            int pos_stem_neighbor = dv->positions()[j];
            std::string stem_neighbor = dv->stems()[pos_stem_neighbor];
            
            std::string uw1("#uw8(");
            std::string uw2("#uw8(");
            uw1 += (stem + ' ' + stem_neighbor + ')');
            uw2 += (stem_neighbor + ' ' + stem + ')');
            
            for(int k=0;k<expressions.size();k++)
            {
                std::string expression = expressions[k];
                if(uw1.compare(stem_expr(expression)) == 0 ||uw2.compare(stem_expr(expression)) == 0)
                {
                    tfs[k]++;
                }
            }
        }
    }

    return tfs;
}

int FeatureGenerator::computeExpressionTf(std::string &expression, indri::api::DocumentVector* dv)
{
    int tf = 0;
    
    for(int i=0;i<dv->positions().size();i++)
    {
        int pos_stem = dv->positions()[i];
        std::string stem = dv->stems()[pos_stem];

        //#1
        if(i<dv->positions().size()-1)
        {
            int pos_stem_after = dv->positions()[i+1];
            std::string stem_after = dv->stems()[pos_stem_after];
            
            std::string phrase("#1(");
            phrase += (stem + ' ' + stem_after + ')');
            if(phrase.compare(stem_expr(expression)) == 0)
            {
                tf++;
            }
        }

        //#uw8
        //range: [j_start, j_end]
        int j_start = i - 8 < 0 ? 0: i - 8;
        int j_end = i + 8 > dv->positions().size() -1 ? dv->positions().size() -1: i + 8;
        
        for(int j = j_start; j<= j_end && j!=i ; j++)
        {
            int pos_stem_neighbor = dv->positions()[j];
            std::string stem_neighbor = dv->stems()[pos_stem_neighbor];
            
            std::string uw1("#uw8(");
            std::string uw2("#uw8(");
            uw1 += (stem + ' ' + stem_neighbor + ')');
            uw2 += (stem_neighbor + ' ' + stem + ')');
            
            if(uw1.compare(stem_expr(expression)) == 0 ||uw2.compare(stem_expr(expression)) == 0)
            {
                tf++;
            }
        }
    }

    return tf;
}

// This one is faster than the previous implementation?
void FeatureGenerator::computeExpressionTfs(std::string &expression, std::vector<lemur::api::DOCID_T> &docids_sorted)
{
    std::string stemmed_expr = stem_expr(expression);
    if(expr_tf_done_set.find(stemmed_expr)!=expr_tf_done_set.end())// Has been computed and stored
        return ;
        
    std::vector<indri::api::ScoredExtentResult> result = env.expressionList(expression);
    int p = 0; // p is the index for pseudo relevant documents
    int q = 0; // q is the index for pos list for expression
    
    while(p < docids_sorted.size() && q < result.size())
    {
        lemur::api::DOCID_T docid_p = docids_sorted[p];
        lemur::api::DOCID_T docid_q = result[q].document;
        
        if(docid_p == docid_q)
        {
            // ==== WARNING: Key here is stemmed expression ====
            std::string newkey = stemmed_expr + "+" + tostr(docid_p);
            int occurence = (int)result[q].score; // occurence will be 1 for each in the result
            expr_doc_tf_map[newkey] += occurence;
            
            q++;
        }
        else if(docid_p > docid_q)
        {
            q++;
        }
        else // docid_p < docid_q
        {
            p++;
        }
    }
    
    std::cout<<"TFS for RAW:<"<<expression<<"> STEMMED:<"<<stemmed_expr<<"> in "<<docids_sorted.size()<<" docs."<<std::endl;

    expr_tf_done_set.insert(stemmed_expr);
}

void FeatureGenerator::load_clicks(std::string &doc_list_file, std::string &snippet_file)
{
    std::vector<std::string> click_docnames;

    std::ifstream click_doc_in(doc_list_file.data(), std::ifstream::in);
    std::string line;
    while(getline(click_doc_in, line))
    {
        click_docnames.push_back(line);
    }
    click_doc_in.close();

    //    if(click_docnames.size() ==0)
    //  return ;
    
    click_docids = env.documentIDsFromMetadata("docno", click_docnames);

    std::ifstream snippet_file_in(snippet_file.data(), std::ifstream::in);
    while(getline(snippet_file_in, line))
    {
        click_snippets.push_back(line);
    }
    snippet_file_in.close();

    // Construct snippet tfidf matrix
    std::vector<Eigen::Triplet<float> > tripleList;
    tripleList.reserve(click_docids.size() * 500);

    int N = env.documentCount();
    
    for(int i=0;i<click_docids.size();i++)
    {
        std::vector<std::string> terms = split(click_snippets[i], ' ');
        std::map<std::string, int> tfs;
        int docLength = terms.size();
        for(int j=0;j<docLength;j++)
        {
            std::string stem = stem_it(terms[j]);
            if(tfs.find(stem)==tfs.end())
                tfs[stem] = 1;
            else
                tfs[stem] = tfs[stem] + 1;
            
        }

        std::map<std::string, int>::iterator it;
        for(it=tfs.begin();it!=tfs.end();++it)
        {
            std::string stem = it->first;
            if(local_stem_dict.find(stem) == local_stem_dict.end())
                continue;

            int stemid = local_stem_dict[stem];
            int tf = it->second;
            int df = env.documentStemCount(stem);
            float tfidf = Utils::TFIDF(tf, docLength, df, N);
            tripleList.push_back(Eigen::Triplet<float>(stemid, i, tfidf));
        }
    }

    click_snippets_matrix = new Eigen::SparseMatrix<float>(local_stem_list.size(), click_docids.size());
    click_snippets_matrix->setFromTriplets(tripleList.begin(), tripleList.end());
}

void FeatureGenerator::reweight_by_snippets()
{
    if(click_snippets_matrix->cols() ==0)
        return ;

    Eigen::SparseMatrix<float> dists_sparse = click_snippets_matrix->transpose() * (*stem_document_matrix);
    Eigen::MatrixXf dists;
    dists = Eigen::MatrixXf(dists_sparse);
    Eigen::VectorXf weights = dists.colwise().sum();
    Eigen::SparseMatrix<float>* new_stem_document_matrix = new Eigen::SparseMatrix<float>(stem_document_matrix->rows(), stem_document_matrix->cols());
    
    for(int i= 0;i<stem_document_matrix->outerSize();i++)
        for(Eigen::SparseMatrix<float>::InnerIterator it(*stem_document_matrix, i); it; ++it)
        {
            int row = it.row();
            int col = it.col();
            float value = it.value();
            new_stem_document_matrix->coeffRef(row, col) = value * (1 + weights[col]);
        }

    delete stem_document_matrix;
    stem_document_matrix = new_stem_document_matrix;
}

void FeatureGenerator::load_qrel(std::string &qrel_file)
{
    std::ifstream qrel_in(qrel_file.c_str(), std::ifstream::in);
    std::string line;
    while(getline(qrel_in, line))
    {
        std::istringstream iss(line);
        int qid;
        std::string temp;
        std::string docname;
        int rel;
        iss>>qid>>temp>>docname>>rel;
        if (qid != queryid)
        {
            continue;
        }

        if(rel > 0)
        {
            qrel_docs.push_back(docname);
        }
        else
        {
            qirrel_docs.push_back(docname);
        }

    }
    
    qrel_in.close();
}

void FeatureGenerator::load_output_terms(std::string &fts_file)
{
    std::ifstream fts_in(fts_file.data(), std::ifstream::in);
    std::string line;
    while(getline(fts_in, line))
    {
        vector<string> parts = split(line, '#');
        vector<string> term_parts = split(parts[parts.size()-1], '\t');
        float true_recall = -1;
        // NOTE: For trec 10 and trec 12, this will be -1. Do not be surprised to see that ;)
        if(parts.size() > 1 && parts[0].compare("")!=0)
        {
            true_recall = atof(split(parts[0], '\t')[0].data());
        }
        int qid = atoi(term_parts[2].data());
        if(qid!=queryid)
            continue;
        output_terms.push_back(term_parts[1]);
        output_recalls[term_parts[1]] = true_recall;
    }
}

void FeatureGenerator::load_stop_words(std::string &stopw_file)
{
    std::ifstream stop_in(stopw_file.data(), std::ifstream::in);
    std::string line;
    while(getline(stop_in, line))
    {
        std::size_t found = line.find("<word>");
        if(found!=std::string::npos)
        {
            std::size_t found2 = line.find("</word>");
            std::string stopw = line.substr(found + 6, found2 - found - 6);
            std::transform(stopw.begin(), stopw.end(), stopw.begin(), ::tolower);
            stop_words.insert(stopw);
        }
        
    }

    // erase us
    stop_words.erase("us");
}

void FeatureGenerator::generateRFMatrix(int num_fbdocs)
{
    qrel_docids = env.documentIDsFromMetadata("docno", qrel_docs);    
    std::vector<Eigen::Triplet<float> > tripleList;
    tripleList.reserve(qrel_docids.size() *  500);

    std::vector<indri::api::DocumentVector*> dvs;

    if(num_fbdocs < qrel_docids.size())
    {
        std::vector<lemur::api::DOCID_T> fb_docids(qrel_docids.begin(), qrel_docids.begin() + num_fbdocs);
        dvs = env.documentVectors(fb_docids);
    }
    else
    {
        dvs = env.documentVectors(qrel_docids);
    }
    
    // Start of local stemid
    int stemid_index = 0;
    int N = env.documentCount();
    //cout<<"DOC VECTOR SIZE:"<<dvs.size()<<endl;
    
    for( int i = 0; i< dvs.size(); i++)
    {
        int docLength = dvs[i]->positions().size();
        int unique_stem_count = dvs[i]->stems().size();
        
        std::vector<int> tfs(unique_stem_count, 0);
        
        for(int j=0;j<dvs[i]->positions().size();j++)
        {
            int pos = dvs[i]->positions()[j];
            tfs[pos]++;
        }
        
        for(int j=0;j<dvs[i]->stems().size();j++)
        {
            std::string stem = dvs[i]->stems()[j];
            int stemid = -1;
            
            // Stem first seen
            if(local_stem_dict.find(stem) == local_stem_dict.end())
            {
                stemid = stemid_index;
                // TODO: Need to check this!
                local_stem_list.push_back(stem);
                local_stem_dict[stem] = stemid_index++;
            }
            else
            {
                stemid = local_stem_dict[stem];
            }
            
            int tf = tfs[j];
            int df = env.documentStemCount(stem);
            float tfidf = Utils::TFIDF(tf, docLength, df, N);
            tripleList.push_back(Eigen::Triplet<float>(stemid, i, tfidf));
        }
        delete dvs[i];
    }
    
    // TFIDF Matrix: Number of stems x Number of docs    
    // Here stemid_index will just be the size of the local vocabulary
    stem_document_matrix = new Eigen::SparseMatrix<float>(stemid_index, dvs.size());
    stem_document_matrix->setFromTriplets(tripleList.begin(), tripleList.end());
}


void FeatureGenerator::generateReweightedTFIDFMatrix(int num_fbdocs)
{
    std::vector<Eigen::Triplet<float> > tripleList;
    tripleList.reserve(res_docids.size() *  500);

    std::vector<indri::api::DocumentVector*> dvs;

    if(num_fbdocs < res_docids.size())
    {
        std::vector<lemur::api::DOCID_T> fb_docids(res_docids.begin(), res_docids.begin() + num_fbdocs);
        dvs = env.documentVectors(fb_docids);
    }
    else
    {
        dvs = env.documentVectors(res_docids);
    }
    
    // Start of local stemid
    int stemid_index = 0;
    int N = env.documentCount();
    //cout<<"DOC VECTOR SIZE:"<<dvs.size()<<endl;
    int sum = dvs.size() * (dvs.size() +1) / 2;
    
    for( int i = 0; i< dvs.size(); i++)
    {
        int docLength = dvs[i]->positions().size();
        int unique_stem_count = dvs[i]->stems().size();
        
        std::vector<int> tfs(unique_stem_count, 0);
        
        for(int j=0;j<dvs[i]->positions().size();j++)
        {
            int pos = dvs[i]->positions()[j];
            tfs[pos]++;
        }

        // Rank weight of document
        //float weight = 1.0 / ( i + 1.0) / (i+1.0) / (i+1.0);
        float weight = ((float)(dvs.size() - i)) / sum;
        
        for(int j=0;j<dvs[i]->stems().size();j++)
        {
            std::string stem = dvs[i]->stems()[j];
            int stemid = -1;
            
            // Stem first seen
            if(local_stem_dict.find(stem) == local_stem_dict.end())
            {
                stemid = stemid_index;
                // TODO: Need to check this!
                local_stem_list.push_back(stem);
                local_stem_dict[stem] = stemid_index++;
            }
            else
            {
                stemid = local_stem_dict[stem];
            }
            
            int tf = tfs[j];
            int df = env.documentStemCount(stem);
            float tfidf = weight * Utils::TFIDF(tf, docLength, df, N);
            tripleList.push_back(Eigen::Triplet<float>(stemid, i, tfidf));
        }
        delete dvs[i];
    }
    
    // TFIDF Matrix: Number of stems x Number of docs    
    // Here stemid_index will just be the size of the local vocabulary
    stem_document_matrix = new Eigen::SparseMatrix<float>(stemid_index, dvs.size());
    stem_document_matrix->setFromTriplets(tripleList.begin(), tripleList.end());
}

void FeatureGenerator::generateTFIDFMatrix(int num_fbdocs)
{
    std::vector<Eigen::Triplet<float> > tripleList;
    tripleList.reserve(res_docids.size() *  500);

    std::vector<Eigen::Triplet<int> > tripleTfList;
    tripleTfList.reserve(res_docids.size() * 500);

    std::vector<indri::api::DocumentVector*> dvs;

    if(num_fbdocs < res_docids.size())
    {
        std::vector<lemur::api::DOCID_T> fb_docids(res_docids.begin(), res_docids.begin() + num_fbdocs);
        dvs = env.documentVectors(fb_docids);
    }
    else
    {
        dvs = env.documentVectors(res_docids);
    }
    
    // Start of local stemid
    int stemid_index = 0;
    int N = env.documentCount();
    //cout<<"DOC VECTOR SIZE:"<<dvs.size()<<endl;
    
    for( int i = 0; i< dvs.size(); i++)
    {
        int docLength = dvs[i]->positions().size();
        int unique_stem_count = dvs[i]->stems().size();
        
        std::vector<int> tfs(unique_stem_count, 0);
        
        for(int j=0;j<dvs[i]->positions().size();j++)
        {
            int pos = dvs[i]->positions()[j];
            tfs[pos]++;
        }

        // store all doc lengths to document_lens
        // [OOV] excluded
        document_lens.push_back(docLength - tfs[0]);

        for(int j=0;j<dvs[i]->stems().size();j++)
        {
            std::string stem = dvs[i]->stems()[j];
            int stemid = -1;
            if(stem.compare("[OOV]")==0)
            {
                continue;
            }
            
            // Stem first seen
            if(local_stem_dict.find(stem) == local_stem_dict.end())
            {
                stemid = stemid_index;
                // TODO: Need to check this!
                local_stem_list.push_back(stem);
                local_stem_dict[stem] = stemid;
                stemid_index++;
            }
            else
            {
                stemid = local_stem_dict[stem];
            }
            
            int tf = tfs[j];
            //Le's imple.
            //if(tf>0)
            //    tf = 1;

            //int df = env.documentCount(stem);
            //if(df==0)
            //    df = 1;
            
            //Mine
            int df = env.documentStemCount(stem);
            float tfidf = Utils::TFIDF(tf, docLength, df, N);
            tripleList.push_back(Eigen::Triplet<float>(stemid, i, tfidf));
            tripleTfList.push_back(Eigen::Triplet<int>(stemid, i, tf));
        }
        delete dvs[i];
    }
    
    // TFIDF Matrix: Number of stems x Number of docs    
    // Here stemid_index will just be the size of the local vocabulary
    stem_document_matrix = new Eigen::SparseMatrix<float>(stemid_index, dvs.size());
    stem_document_matrix->setFromTriplets(tripleList.begin(), tripleList.end());

    stem_document_tf_matrix = new Eigen::SparseMatrix<int>(stemid_index, dvs.size());
    stem_document_tf_matrix->setFromTriplets(tripleTfList.begin(), tripleTfList.end());
}

void FeatureGenerator::generateTFIDFMatrixWithPhrase(int num_fbdocs)
{
    prf_total_length = 0;
    std::vector<Eigen::Triplet<float> > tripleList;
    tripleList.reserve(res_docids.size() *  500);

    std::vector<lemur::api::DOCID_T> *fb_docids;

    if(num_fbdocs < res_docids.size())
    {
        fb_docids = new std::vector<lemur::api::DOCID_T>(res_docids.begin(), res_docids.begin() + num_fbdocs);
    }
    else
    {
        fb_docids = new std::vector<lemur::api::DOCID_T>(res_docids);
    }

    std::vector<indri::api::DocumentVector*> dvs = env.documentVectors(*fb_docids);
    
    // Start of local stemid
    int stemid_index = 0;
    int N = env.documentCount();
    //cout<<"DOC VECTOR SIZE:"<<dvs.size()<<endl;

    double _tf_computing_time_ = 0;

    // This sorting is important for getting tfs for expressions!
    std::vector<lemur::api::DOCID_T> fb_docids_sorted(*fb_docids);
    std::sort(fb_docids_sorted.begin(), fb_docids_sorted.end());

    // Add qrel docs and get TFS for both qrel docs and prf docs
    // This is important!
    // Otherwise, TFs for expressions will be bugged!!!!!!!
    qrel_docids = env.documentIDsFromMetadata("docno", qrel_docs);
    std::cout<<"qrel docs:"<<qrel_docids.size()<<std::endl;
    std::set<lemur::api::DOCID_T> all_docids_set;
    for(int i=0;i<qrel_docids.size();i++)
    {
        all_docids_set.insert(qrel_docids[i]);
    }
    
    for(int i=0;i<fb_docids->size();i++)
    {
        all_docids_set.insert((*fb_docids)[i]);
    }

    std::vector<lemur::api::DOCID_T> all_docids(all_docids_set.begin(), all_docids_set.end());
    std::sort(all_docids.begin(), all_docids.end());
    
    for( int i = 0; i< dvs.size(); i++)
    {
        int docLength = env.documentLength((*fb_docids)[i]);
        prf_total_length += docLength;
        int unique_stem_count = dvs[i]->stems().size();
        
        std::vector<int> tfs(unique_stem_count, 0);
        
        for(int j=0;j<dvs[i]->positions().size();j++)
        {
            int pos_stem = dvs[i]->positions()[j];
            tfs[pos_stem]++;
        }

        // store all doc lengths to document_lens
        document_lens.push_back(docLength);

        // Compute tfidf entries for individual stems
        for(int j=0;j<dvs[i]->stems().size();j++)
        {
            std::string stem = dvs[i]->stems()[j];
            int stemid = -1;
            if(stem.compare("[OOV]")==0)
            {
                continue;
            }

            int tf = tfs[j];
            //Le's imple.
            //if(tf>0)
            //    tf = 1;

            // Accumulate tf for prf documents
            if(expr_prf_tf_map.find(stem)==expr_prf_tf_map.end())
                expr_prf_tf_map[stem] = 0;
            else
                expr_prf_tf_map[stem] += tf;

            int df = env.documentStemCount(stem);
            float tfidf = Utils::TFIDF(tf, docLength, df, N);
            if(tfidf == 0)
                continue;
            
            // Stem first seen
            if(local_stem_dict.find(stem) == local_stem_dict.end())
            {
                stemid = stemid_index;
                // TODO: Need to check this!
                local_stem_list.push_back(stem);
                local_stem_dict[stem] = stemid;
                stemid_index++;
            }
            else
            {
                stemid = local_stem_dict[stem];
            }
            
            tripleList.push_back(Eigen::Triplet<float>(stemid, i, tfidf));
        }

        // Compute tfidf entries for compound expressions (phrases, etc.)
        //std::vector<int> phrase_tfs = computeExpressionTfs(output_phrases, dvs[i]);

        for(int j=0;j<output_phrases_unique.size();j++)
        {
            std::string expr = output_phrases_unique[j];
            std::string stemmed_expr = stem_expr(expr);

            // expr must in indri query language format
            // Method 1, is a bit slow

            double _tf_compute_start_ = omp_get_wtime();
            //           computeExpressionTfs(expr, fb_docids_sorted);
            computeExpressionTfs(expr, all_docids);
            double _tf_compute_end_ = omp_get_wtime();
            _tf_computing_time_ += (_tf_compute_end_ - _tf_compute_start_);

            int tf = expr_doc_tf_map[stemmed_expr + "+" + tostr((*fb_docids)[i])];

            // Accumulate tf for prf documents
            if(expr_prf_tf_map.find(stemmed_expr)==expr_prf_tf_map.end())
                expr_prf_tf_map[stemmed_expr] = 0;
            else
                expr_prf_tf_map[stemmed_expr] += tf;

            
            // Method 2, 
            //int tf = computeExpressionTf(expr, dvs[i]);
            //int tf = phrase_tfs[j];


            int df = env.documentExpressionCount( expr );
            float tfidf = Utils::TFIDF(tf, docLength, df, N);
            if(df == 0 || tfidf == 0 )
                continue;
            
            int stemid = -1;
            
            if(local_stem_dict.find(stemmed_expr) == local_stem_dict.end())
            {
                stemid = stemid_index;
                // TODO: Need to check this!
                local_stem_list.push_back(stemmed_expr);
                local_stem_dict[stemmed_expr] = stemid;
                stemid_index++;
            }
            else
            {
                stemid = local_stem_dict[stemmed_expr];
            }

            tripleList.push_back(Eigen::Triplet<float>(stemid, i, tfidf));
        }
        
        delete dvs[i];
    }

    std::cout<<"Expression Tfs computation time: "<< _tf_computing_time_ <<" sec."<<std::endl;

    delete fb_docids;
    
    // TFIDF Matrix: Number of stems x Number of docs    
    // Here stemid_index will just be the size of the local vocabulary
    stem_document_matrix = new Eigen::SparseMatrix<float>(stemid_index, dvs.size());
    stem_document_matrix->setFromTriplets(tripleList.begin(), tripleList.end());
}

// This implementation contains top bigrams and un-ordered windows from top documents
void  FeatureGenerator::generateTFIDFMatrixWithExpressions(int num_fbdocs)
{
    prf_total_length = 0;
    std::vector<Eigen::Triplet<float> > tripleList;
    tripleList.reserve(res_docids.size() *  500);

    std::vector<lemur::api::DOCID_T> *fb_docids;

    if(num_fbdocs < res_docids.size())
    {
        fb_docids = new std::vector<lemur::api::DOCID_T>(res_docids.begin(), res_docids.begin() + num_fbdocs);
    }
    else
    {
        fb_docids = new std::vector<lemur::api::DOCID_T>(res_docids);
    }

    std::vector<indri::api::DocumentVector*> dvs = env.documentVectors(*fb_docids);
    
    // Start of local stemid
    int stemid_index = 0;
    int N = env.documentCount();
    //cout<<"DOC VECTOR SIZE:"<<dvs.size()<<endl;

    double _tf_computing_time_ = 0;

    // This sorting is important for getting tfs for expressions!
    std::vector<lemur::api::DOCID_T> fb_docids_sorted(*fb_docids);
    std::sort(fb_docids_sorted.begin(), fb_docids_sorted.end());

    std::set<std::string> collection_stemmed_bigram_set, collection_stemmed_uw_set;
    
    for( int i = 0; i< dvs.size(); i++)
    {
        int docLength = env.documentLength((*fb_docids)[i]);
        prf_total_length += docLength;
        int unique_stem_count = dvs[i]->stems().size();
        
        std::vector<int> tfs(unique_stem_count, 0);
        
        for(int j=0;j<dvs[i]->positions().size();j++)
        {
            int pos_stem = dvs[i]->positions()[j];
            tfs[pos_stem]++;
            
            // Collection bi-grams and unordered windows
            if(j+1<dvs[i]->positions().size())
            {
                int pos_stem_next = dvs[i]->positions()[j+1];
                std::string stem = dvs[i]->stems()[pos_stem];
                std::string stem_next = dvs[i]->stems()[pos_stem_next];
                
                if(stem.compare("[OOV]")!=0 && stem_next.compare("[OOV]")!=0)
                {
                    std::string bigram = "#1("+stem+" "+stem_next+")";
                    std::string uw = "#uw8("+stem+" "+stem_next+")";
                    
                    // This IF and the following is very IMPORTANT!!!
                    if(output_phrase_set.find(bigram)==output_phrase_set.end())
                    {
                        collection_stemmed_bigram_set.insert(bigram);
                    }

                    if(output_phrase_set.find(uw)==output_phrase_set.end())
                    {
                        collection_stemmed_uw_set.insert(uw);
                    }
                }
            }
        }

        // store all doc lengths to document_lens
        document_lens.push_back(docLength);

        // Compute tfidf entries for individual stems
        for(int j=0;j<dvs[i]->stems().size();j++)
        {
            std::string stem = dvs[i]->stems()[j];
            int stemid = -1;
            if(stem.compare("[OOV]")==0)
            {
                continue;
            }

            int tf = tfs[j];
            //Le's imple.
            //if(tf>0)
            //    tf = 1;

            // Accumulate tf for prf documents
            if(expr_prf_tf_map.find(stem)==expr_prf_tf_map.end())
                expr_prf_tf_map[stem] = 0;
            else
                expr_prf_tf_map[stem] += tf;

            int df = env.documentStemCount(stem);
            float tfidf = Utils::TFIDF(tf, docLength, df, N);
            if(tfidf == 0)
                continue;
            
            // Stem first seen
            if(local_stem_dict.find(stem) == local_stem_dict.end())
            {
                stemid = stemid_index;
                // TODO: Need to check this!
                local_stem_list.push_back(stem);
                local_stem_dict[stem] = stemid;
                stemid_index++;
            }
            else
            {
                stemid = local_stem_dict[stem];
            }
            
            tripleList.push_back(Eigen::Triplet<float>(stemid, i, tfidf));
        }

    }

    // As the number of bigrams and uw in collection is too large
    // Sample from them !!!
    int collection_stem_count = stemid_index;
    std::set<std::string> sampled_stemmed_expr_set;

    int sample_target = collection_stem_count * 0.5;
    int sample_count = 0;
    for(std::set<std::string>::iterator it=collection_stemmed_bigram_set.begin(); it!=collection_stemmed_bigram_set.end();++it)
    {
        if(collection_stemmed_bigram_set.size() < sample_target)
            collection_stemmed_expr_set.insert(*it);
        else
        {
            double ratio = (double) sample_target / collection_stemmed_bigram_set.size();
            double r = (double) rand() / RAND_MAX;
            if(r <= ratio) // Sample it!
            {
                collection_stemmed_expr_set.insert(*it);
                sample_count ++;
            }
            
            if(sample_count >= sample_target)
                break;
        }
    }       

    sample_count = 0;
    for(std::set<std::string>::iterator it=collection_stemmed_uw_set.begin(); it!=collection_stemmed_uw_set.end();++it)
    {
        if(collection_stemmed_uw_set.size() < sample_target)
            collection_stemmed_expr_set.insert(*it);
        else
        {
            double ratio = (double) sample_target / collection_stemmed_uw_set.size();
            double r = (double) rand() / RAND_MAX;
            if(r <= ratio) // Sample it!
            {
                collection_stemmed_expr_set.insert(*it);
                sample_count ++;
            }
            
            if(sample_count >= sample_target)
                break;
        }
    }       
    // Sample done

    // Add exprs in queries into set of exprs
    for(int i=0;i<output_phrases.size();i++)
    {
        std::string expr = output_phrases[i];
        collection_stemmed_expr_set.insert(expr); // NOTE HERE collection_stemmed_expr contains all stemmed_expr, 
                                                  // but this expr is not, INTENTIONALLY SO!!!
    }
    
    std::cout<<"Collection stem count: "<< stemid_index << std::endl;
    std::cout<<"Collection expr count: "<< collection_stemmed_expr_set.size()<<std::endl;

    // Compute tfidf entries for compound expressions (phrases, etc.)
    //std::vector<int> phrase_tfs = computeExpressionTfs(output_phrases, dvs[i]);
    for(std::set<std::string>::iterator it=collection_stemmed_expr_set.begin(); it!=collection_stemmed_expr_set.end();++it)
        // for(int j=0;j<output_phrases.size();j++)
    {
        std::string expr = *it;
        
        // expr must in indri query language format
        // Method 1, is a bit slow
        
        double _tf_compute_start_ = omp_get_wtime();
        computeExpressionTfs(expr, fb_docids_sorted);
        double _tf_compute_end_ = omp_get_wtime();
        _tf_computing_time_ += (_tf_compute_end_ - _tf_compute_start_);

        expr_df_map[expr] = env.documentExpressionCount( expr );
    }

    std::cout<<"Expression Tfs computation time: "<< _tf_computing_time_ <<" sec."<<std::endl;

    for(int i=0;i<dvs.size();i++)
    {
        int docLength = env.documentLength((*fb_docids)[i]);
        for(std::set<std::string>::iterator it=collection_stemmed_expr_set.begin(); it!=collection_stemmed_expr_set.end();++it)
        {
            std::string expr = *it;
            int tf = expr_doc_tf_map[expr + "+" + tostr((*fb_docids)[i])];
        
            // Accumulate tf for prf documents
            if(expr_prf_tf_map.find(expr)==expr_prf_tf_map.end())
                expr_prf_tf_map[expr] = 0;
            else
                expr_prf_tf_map[expr] += tf;
            
            int df = expr_df_map[expr];
            float tfidf = Utils::TFIDF(tf, docLength, df, N);
            if(df == 0 || tfidf == 0 )
                continue;
            
            int stemid = -1;
            
            if(local_stem_dict.find(expr) == local_stem_dict.end())
            {
                stemid = stemid_index;
                // TODO: Need to check this!
                local_stem_list.push_back(expr);
                local_stem_dict[expr] = stemid;
                stemid_index++;
            }
            else
            {
                stemid = local_stem_dict[expr];
            }

            tripleList.push_back(Eigen::Triplet<float>(stemid, i, tfidf));
        }
        
        delete dvs[i];
    }

    delete fb_docids;
    
    // TFIDF Matrix: Number of stems x Number of docs    
    // Here stemid_index will just be the size of the local vocabulary
    stem_document_matrix = new Eigen::SparseMatrix<float>(stemid_index, dvs.size());
    stem_document_matrix->setFromTriplets(tripleList.begin(), tripleList.end());
    std::cout<<"TFIDF matrix computation done."<<std::endl;
}

// Save stem list and TFIDF matrix to prefix.tfidf and prefix.stem
void FeatureGenerator::saveTFIDFDataMatrix(std::string &prefix)
{
    std::string tfidf_file = prefix + ".tfidf";
    std::string stem_file = prefix + ".stem";
    std::ofstream data_out(tfidf_file.data(), std::ofstream::out);
    std::ofstream term_out(stem_file.data(), std::ofstream::out);
    
    // Save terms
    std::map<std::string, int>::iterator iter;

    for (iter = local_stem_dict.begin(); iter != local_stem_dict.end(); ++iter) 
    {
        term_out<<iter->first<<" "<<(iter->second+1)<<std::endl;
    }
    term_out.close();

    // Save data sparse format
    /*
    for(int i=0;i<stem_document_matrix->rows();i++)
    {
        for(int j=0;j<stem_document_matrix->cols();j++)
        {
            float value = stem_document_matrix->coeffRef(i,j);
            if(value!=0)
            {
                data_out<<(i+1)<<" "<<(j+1)<<" "<<value<<std::endl;
            }
        }
    }
    */

    // Save data dense format
    for(int i=0;i<stem_document_matrix->rows();i++)
    {
        for(int j=0;j<stem_document_matrix->cols();j++)
        {
            float value = stem_document_matrix->coeffRef(i,j);
            data_out<<value<<" ";
        }
        data_out<<endl;
    }

    data_out.close();
}

void FeatureGenerator::generateRandomTFIDFMatrix(int num_fbdocs)
{
    std::vector<Eigen::Triplet<float> > tripleList;
    tripleList.reserve(res_docids.size() *  500);

    std::vector<indri::api::DocumentVector*> dvs;
    std::vector<lemur::api::DOCID_T> fb_docids;
    int fbdoc_size = res_docids.size()<num_fbdocs?res_docids.size():num_fbdocs;
    
    
    // Uniform sampling START
    /*
    srand(time(NULL));
    for(int i=0;i<res_docids.size();i++)
    {
        int rand_index = rand() % res_docids.size();
        fb_docids.push_back(res_docids[rand_index]);
        }*/
    // Uniform sampling END

    // Biased sampling START

    BiasedSampler2 bs(fbdoc_size);
    for(int i=0;i<fbdoc_size;i++)
    {
        int rand_index = bs.getSample();
        fb_docids.push_back(res_docids[rand_index]);
    }
    // Biased sampling END
    
    dvs = env.documentVectors(fb_docids);
    
    // Start of local stemid
    int stemid_index = 0;
    int N = env.documentCount();
    //cout<<"DOC VECTOR SIZE:"<<dvs.size()<<endl;
    
    for( int i = 0; i< dvs.size(); i++)
    {
        int docLength = dvs[i]->positions().size();
        int unique_stem_count = dvs[i]->stems().size();
        
        std::vector<int> tfs(unique_stem_count, 0);
        
        for(int j=0;j<dvs[i]->positions().size();j++)
        {
            int pos = dvs[i]->positions()[j];
            tfs[pos]++;
        }
        
        for(int j=0;j<dvs[i]->stems().size();j++)
        {
            std::string stem = dvs[i]->stems()[j];
            int stemid = -1;
            
            // Stem first seen
            if(local_stem_dict.find(stem) == local_stem_dict.end())
            {
                stemid = stemid_index;
                // TODO: Need to check this!
                local_stem_list.push_back(stem);
                local_stem_dict[stem] = stemid_index++;
            }
            else
            {
                stemid = local_stem_dict[stem];
            }
            
            int tf = tfs[j];
            int df = env.documentStemCount(stem);
            float tfidf = Utils::TFIDF(tf, docLength, df, N);
            tripleList.push_back(Eigen::Triplet<float>(stemid, i, tfidf));
        }
        delete dvs[i];
    }
    
    // TFIDF Matrix: Number of stems x Number of docs    
    // Here stemid_index will just be the size of the local vocabulary
    stem_document_matrix = new Eigen::SparseMatrix<float>(stemid_index, dvs.size());
    stem_document_matrix->setFromTriplets(tripleList.begin(), tripleList.end());
}

void FeatureGenerator::generateRandomReweightedTFIDFMatrix(int num_fbdocs)
{
    std::vector<Eigen::Triplet<float> > tripleList;
    tripleList.reserve(res_docids.size() *  500);

    std::vector<indri::api::DocumentVector*> dvs;
    std::vector<lemur::api::DOCID_T> fb_docids;
    int fbdoc_size = res_docids.size()<num_fbdocs?res_docids.size():num_fbdocs;

    int sum = fbdoc_size * (fbdoc_size + 1) / 2;
    
    // Uniform sampling START
    /*
    srand(time(NULL));
    for(int i=0;i<res_docids.size();i++)
    {
        int rand_index = rand() % res_docids.size();
        fb_docids.push_back(res_docids[rand_index]);
        }*/
    // Uniform sampling END

    // Biased sampling START


    BiasedSampler2 bs(fbdoc_size);
    std::vector<int> sample_index;
    for(int i=0;i<fbdoc_size;i++)
    {
        int rand_index = bs.getSample();
        sample_index.push_back(rand_index);
        fb_docids.push_back(res_docids[rand_index]);
    }
    // Biased sampling END
    
    dvs = env.documentVectors(fb_docids);
    
    // Start of local stemid
    int stemid_index = 0;
    int N = env.documentCount();
    //cout<<"DOC VECTOR SIZE:"<<dvs.size()<<endl;

    
    for( int i = 0; i< dvs.size(); i++)
    {
        int docLength = dvs[i]->positions().size();
        int unique_stem_count = dvs[i]->stems().size();
        
        std::vector<int> tfs(unique_stem_count, 0);
        
        for(int j=0;j<dvs[i]->positions().size();j++)
        {
            int pos = dvs[i]->positions()[j];
            tfs[pos]++;
        }

        float weight = ((float)(dvs.size() - sample_index[i])) / sum;
        
        for(int j=0;j<dvs[i]->stems().size();j++)
        {
            std::string stem = dvs[i]->stems()[j];
            int stemid = -1;
            
            // Stem first seen
            if(local_stem_dict.find(stem) == local_stem_dict.end())
            {
                stemid = stemid_index;
                // TODO: Need to check this!
                local_stem_list.push_back(stem);
                local_stem_dict[stem] = stemid_index++;
            }
            else
            {
                stemid = local_stem_dict[stem];
            }
            
            int tf = tfs[j];
            int df = env.documentStemCount(stem);
            float tfidf = weight * Utils::TFIDF(tf, docLength, df, N);
            tripleList.push_back(Eigen::Triplet<float>(stemid, i, tfidf));
        }
        delete dvs[i];
    }
    
    // TFIDF Matrix: Number of stems x Number of docs    
    // Here stemid_index will just be the size of the local vocabulary
    stem_document_matrix = new Eigen::SparseMatrix<float>(stemid_index, dvs.size());
    stem_document_matrix->setFromTriplets(tripleList.begin(), tripleList.end());
}

void FeatureGenerator::compute_concept_plsa_matrix(int K)
{
    PLSA plsa(stem_document_matrix->transpose(), K);
    plsa.initialize();
    int max_iter = 1;
    plsa.train(max_iter);
    stem_concept_plsa_matrix = plsa.getTheta().transpose();
}

// K is the number of latent topics to keep
void FeatureGenerator::compute_concept_matrix(int K)
{
    if(svd == NULL)
    {
        //        cout<<"Computing SVD..."<<endl;
        svd = new Eigen::JacobiSVD<Eigen::MatrixXf>(*stem_document_matrix, Eigen::ComputeThinU);
        Eigen::MatrixXf U = svd->matrixU();
        //Eigen::MatrixXf S = svd->singularValues().head(K).asDiagonal();

        // Original: do not multiply with singular values
        stem_concept_matrix = (U.cols() > K)?U.leftCols(K):U;

        // Multiply with singular values
        //stem_concept_matrix = ((U.cols() > K)?U.leftCols(K):U) * S ;
    }
}

float FeatureGenerator::concept_distance(std::string &stem1, std::string &stem2, const Eigen::MatrixXf& scm)
{
    // Need to define this according to the distance
    float least_possible_distance = 0;
    
    if(local_stem_dict.find(stem1) == local_stem_dict.end() || local_stem_dict.find(stem2) == local_stem_dict.end())
    {
        return least_possible_distance;
    }
    
    int tid1 = local_stem_dict[stem1];
    int tid2 = local_stem_dict[stem2];
    
    return concept_distance_by_tid(tid1, tid2, scm);
}

std::string FeatureGenerator::concept_neighbor(std::string &stem, int kth, const Eigen::MatrixXf& scm) // kth >=1
{
    if(local_stem_dict.find(stem) == local_stem_dict.end())
    {
        return "";
    }
    
    int tid = local_stem_dict[stem];
    
    if(stem_topk_map.find(tid) == stem_topk_map.end())
    {
        Eigen::VectorXf dists = scm * scm.row(tid).transpose();
        
        // Cosine
        //dists /= stem_concept_matrix.row(tid).norm();
        //dists = (dists.array()/stem_concept_matrix.rowwise().norm().array()).matrix();

        stem_topk_map[tid] = argsort(dists);
    }
    
    // HERE: kth >= 1
    return local_stem_list[stem_topk_map[tid](kth-1)] ;    
}

std::string FeatureGenerator::concept_query_neighbor(std::string &stem, int kth, const Eigen::MatrixXf& scm) // kth >= 1
{
    if(local_stem_dict.find(stem) == local_stem_dict.end())
    {
        return "";
    }
    
    int tid = local_stem_dict[stem];
    
    if(stem_topk_map.find(tid) == stem_topk_map.end())
    {
        Eigen::VectorXf dists = scm * scm.row(tid).transpose();
        
        // Cosine
        //dists /= stem_concept_matrix.row(tid).norm();
        //dists = (dists.array()/stem_concept_matrix.rowwise().norm().array()).matrix();

        stem_topk_map[tid] = argsort(dists);
    }

    int inquery_index = 0;

    for(int i=0;i<stem_topk_map[tid].size();i++)
    {
        std::string query_stem = local_stem_list[stem_topk_map[tid](i)];
        
        //        if(query_stem.compare(stem)!=0 && (output_phrase_set.find(query_stem)!=output_phrase_set.end()||output_stem_set.find(query_stem)!=output_stem_set.end()))
        if(output_phrase_set.find(query_stem)!=output_phrase_set.end()||output_stem_set.find(query_stem)!=output_stem_set.end())
        {
            inquery_index ++;
            if(inquery_index == kth)
                return query_stem;
        }
    }

    return "";
}

float FeatureGenerator::concept_distance(std::string &stem, int kth, const Eigen::MatrixXf& scm) // kth >=1
{
    float least_possible_distance = 0;
    
    if(local_stem_dict.find(stem) == local_stem_dict.end())
    {
        return least_possible_distance;
    }
    
    int tid = local_stem_dict[stem];
    
    if(stem_topk_map.find(tid) == stem_topk_map.end())
    {
        Eigen::VectorXf dists = scm * scm.row(tid).transpose();
        
        //Cosine
        //dists /= stem_concept_matrix.row(tid).norm();
        //dists = (dists.array()/stem_concept_matrix.rowwise().norm().array()).matrix();

        stem_topk_map[tid] = argsort(dists);
    }
    
    // HERE: kth >= 1
    return concept_distance_by_tid(tid, stem_topk_map[tid](kth-1), scm);
}

float FeatureGenerator::concept_distance_by_tid(int tid1, int tid2, const Eigen::MatrixXf& scm)
{
    Eigen::VectorXf v1 = scm.row(tid1);
    Eigen::VectorXf v2 = scm.row(tid2);

    float snormi = v1.squaredNorm();

    if(snormi<0.0001) 
        snormi = 1;
    return v1.dot(v2)/sqrt(snormi);

    // Cosine
    //return (v1.dot(v2)/v1.norm())/v2.norm();
    return v1.dot(v2);// My original solution
}

float FeatureGenerator::compute_pseudo_recall(std::string &stem)
{
    std::vector<indri::api::DocumentVector*> dvs = env.documentVectors(res_docids);
    int pr_size = stem_document_matrix->cols();

    if(pr_size > dvs.size())
        pr_size = dvs.size();

    float has_stem = 0;
    for(int i=0; i< pr_size; i++)
    {
        for(int j=0;j<dvs[i]->stems().size();j++)
        {
            if(stem.compare(dvs[i]->stems()[j])==0) //Equal
            {
                has_stem += 1.0;
                break;
            }
            
            /*
              if(std::find(dvs[i]->stems().begin(), dvs[i]->stems().end(), stem)!=dvs[i]->stems().end())
            {
                has_stem += 1.0;
                }
            */       
        }
        delete dvs[i];
    }    

    //std::cout<<has_stem<<" "<<docids.size()<<std::endl;
    
    return  ( has_stem + 1 )/( pr_size + 2 );// Laplacian smoothing
}

float FeatureGenerator::compute_pseudo_recall_tf(std::string &stem)
{
    std::vector<indri::api::DocumentVector*> dvs = env.documentVectors(res_docids);
    int pr_size = stem_document_matrix->cols();

    if(pr_size > dvs.size())
        pr_size = dvs.size();

    float total_words = 0;
    float tf = 0;

    for(int i=0; i< pr_size; i++)
    {
        total_words += dvs[i]->stems().size();
        for(int j=0;j<dvs[i]->stems().size();j++)
        {
            if(stem.compare(dvs[i]->stems()[j])==0) //Equal
            {
                tf += 1.0;
            }
            
            /*
              if(std::find(dvs[i]->stems().begin(), dvs[i]->stems().end(), stem)!=dvs[i]->stems().end())
            {
                has_stem += 1.0;
                }
            */       
        }
        delete dvs[i];
    }    

    //std::cout<<has_stem<<" "<<docids.size()<<std::endl;
    
    return  ( tf + 1 )/( total_words + 2 );// Laplacian smoothing
}

float FeatureGenerator::compute_weighted_pseudo_recall(std::string &stem)
{
    std::vector<indri::api::DocumentVector*> dvs = env.documentVectors(res_docids);
    int pr_size = stem_document_matrix->cols();

    if(pr_size > dvs.size())
        pr_size = dvs.size();

    float has_stem = 0;
    float sum_score = 0;
    for(int i=0; i< pr_size; i++)
    {
        for(int j=0;j<dvs[i]->stems().size();j++)
        {
            if(stem.compare(dvs[i]->stems()[j])==0) //Equal
            {
                has_stem += exp(res_scores[i]);
                break;
            }

        }
        delete dvs[i];
        sum_score += exp(res_scores[i]);
    }    

    //std::cout<<has_stem<<" "<<docids.size()<<std::endl;
    return  has_stem / sum_score;
    //return  ( has_stem + 1 )/( sum_score + 2 );// Laplacian smoothing
}

float FeatureGenerator::compute_weighted_pseudo_recall_tf(std::string &stem)
{
    std::vector<indri::api::DocumentVector*> dvs = env.documentVectors(res_docids);
    int pr_size = stem_document_matrix->cols();

    if(pr_size > dvs.size())
        pr_size = dvs.size();

    float weighted_total_words = 0;
    float weighted_tf = 0;
    

    for(int i=0; i< pr_size; i++)
    {
        float tf = 0;
        weighted_total_words += exp(res_scores[i]) *dvs[i]->stems().size();
        for(int j=0;j<dvs[i]->stems().size();j++)
        {
            if(stem.compare(dvs[i]->stems()[j])==0) //Equal
            {
                tf += 1.0;
            }

        }

        weighted_tf += tf * exp(res_scores[i]);
        delete dvs[i];
    }    

    //std::cout<<has_stem<<" "<<docids.size()<<std::endl;
    return  weighted_tf / weighted_total_words;
    //return  ( has_stem + 1 )/( sum_score + 2 );// Laplacian smoothing
}

float FeatureGenerator::compute_guided_pseudo_recall(std::string &stem, int n_rel,  std::vector<indri::api::DocumentVector*> & rel_seed_dvs,  Eigen::SparseMatrix<float>* &rel_seeds_stem_document_matrix)
{
    std::vector<indri::api::DocumentVector*> dvs = env.documentVectors(res_docids);    
    int pr_size = stem_document_matrix->cols();

    if(pr_size > dvs.size())
        pr_size = dvs.size();

    //compute similarity to all relevant seeds
    std::vector<argsort_pair> sims_to_rel_seeds(pr_size);

    for(int i=0;i<pr_size;i++)
    {
        argsort_pair item(i, -1);
        Eigen::VectorXf di = Eigen::VectorXf(stem_document_matrix->innerVector(i));
        for(int j=0;j<rel_seed_dvs.size();j++)
        {
            Eigen::VectorXf sj = Eigen::VectorXf(rel_seeds_stem_document_matrix->innerVector(j)).head(di.size());
            float sim = di.dot(sj);
            sim /= di.norm();
            sim /= sj.norm();
            if(sim>item.second)
            {
                item.second = sim;
            }
        }

        sims_to_rel_seeds.push_back(item);
    }

    std::sort(sims_to_rel_seeds.begin(), sims_to_rel_seeds.end(), argsort_comp);
    

    // Compute recall
    float has_stem = 0;
    float sum_score = 0;
    for(int i=0; i< pr_size && i<n_top; i++)
    {
        for(int j=0;j<dvs[sims_to_rel_seeds[i].first]->stems().size();j++)
        {
            if(stem.compare(dvs[sims_to_rel_seeds[i].first]->stems()[j])==0) //Equal
            {
                has_stem += sims_to_rel_seeds[i].second;
                break;
            }

        }
        sum_score += sims_to_rel_seeds[i].second;
    }    

    for(int i=0;i<pr_size;i++)
    {
        delete dvs[i];
    }

    //std::cout<<has_stem<<" "<<docids.size()<<std::endl;
    return  has_stem / sum_score;
    //return  ( has_stem + 1 )/( sum_score + 2 );// Laplacian smoothing
}


float FeatureGenerator::compute_true_index_recall(std::string &stem, const std::vector<std::string> &query_stems)
{
    std::vector<lemur::api::DOCID_T> docids = env.documentIDsFromMetadata("docno", qrel_docs);
    std::vector<indri::api::DocumentVector*> dvs = env.documentVectors(docids);

    float has_stem = 0;
    float index_doc = 0;
    
    for(int i=0; i<dvs.size();i++)
    {
        if(std::find(dvs[i]->stems().begin(), dvs[i]->stems().end(), stem)!=dvs[i]->stems().end())
        {
            has_stem += 1.0;
            index_doc += 1.0;
        }
        else
        {
            for(int j=0;j<query_stems.size();j++)
            {
                if(std::find(dvs[i]->stems().begin(), dvs[i]->stems().end(), query_stems[j])!=dvs[i]->stems().end())
                {
                    index_doc += 1.0;
                    break;
                }
            }
        }
        
        
        delete dvs[i];
    }    

    //std::cout<<has_stem<<" "<<docids.size()<<std::endl;
    if(index_doc==0)
        return 0.5;

    return   has_stem / index_doc;// Laplacian smoothing
    //return  ( has_stem + 1)/(index_doc + 2);// Laplacian smoothing    
}

void FeatureGenerator::get_query_stems(std::vector<std::string> &query_stems)
{
    std::set<std::string>::iterator it;
    for (it=query_terms.begin() ; it != query_terms.end(); it++ )
    {
        std::string term = *it;
        std::string stem = stem_it(term);
        query_stems.push_back(stem);
    }
}

float FeatureGenerator::compute_expr_true_recall(std::string& expr)
{
    std::string stemmed_expr = stem_expr(expr);

    std::vector<lemur::api::DOCID_T> docids = env.documentIDsFromMetadata("docno", qrel_docs);
    std::vector<lemur::api::DOCID_T> docids_sorted(docids);
    std::sort(docids_sorted.begin(), docids_sorted.end());
    computeExpressionTfs(expr, docids_sorted);

    float has_expr = 0;
    
    for(int i=0;i<docids.size();i++)
    {
        int tf = expr_doc_tf_map[stemmed_expr + "+" + tostr(docids[i])];
        if(tf>0)
            has_expr += 1.0;
    }

    return  ( has_expr + 1.0 ) / (docids.size() + 2.0);// Laplacian smoothing
}

float FeatureGenerator::compute_true_recall(std::string &stem)
{
    // Intenionally stem twice!!!
    //std::string stem = stem_it(sstem);
    
    std::vector<lemur::api::DOCID_T> docids = env.documentIDsFromMetadata("docno", qrel_docs);
    std::vector<indri::api::DocumentVector*> dvs = env.documentVectors(docids);

    float has_stem = 0;
    for(int i=0; i<dvs.size();i++)
    {
        for(int j=0;j<dvs[i]->stems().size();j++)
        {
            if(stem.compare(dvs[i]->stems()[j])==0) //Equal
            {
                has_stem += 1.0;
                break;
            }
            
            /*
              if(std::find(dvs[i]->stems().begin(), dvs[i]->stems().end(), stem)!=dvs[i]->stems().end())
            {
                has_stem += 1.0;
                }
            */       
        }
        delete dvs[i];
    }    

    //std::cout<<stem<<":"<<has_stem<<" "<<docids.size()<<std::endl;
    
    return  ( has_stem + 1.0)/(docids.size() + 2.0);// Laplacian smoothing
}

float FeatureGenerator::compute_true_recall_tf(std::string &stem)
{
    std::vector<lemur::api::DOCID_T> docids = env.documentIDsFromMetadata("docno", qrel_docs);
    std::vector<indri::api::DocumentVector*> dvs = env.documentVectors(docids);

    float total_words = 0;
    float tf = 0;

    for(int i=0; i<dvs.size();i++)
    {
        total_words += dvs[i]->stems().size();
        
        for(int j=0;j<dvs[i]->stems().size();j++)
        {
            if(stem.compare(dvs[i]->stems()[j])==0) //Equal
            {
                tf += 1.0;// Bug, this is traversing all stems only once, so tf will always <= 1
            }
            
            /*
              if(std::find(dvs[i]->stems().begin(), dvs[i]->stems().end(), stem)!=dvs[i]->stems().end())
            {
                has_stem += 1.0;
                }
            */       
        }
        delete dvs[i];
    }    

    //std::cout<<has_stem<<" "<<docids.size()<<std::endl;

    return (tf  + 1)/ (total_words + 2); // Laplacian smoothing
}

float FeatureGenerator::compute_rocchio_weight(std::string &stem, float lambda)
{
    std::vector<indri::api::DocumentVector*> dvs = env.documentVectors(res_docids);
    int pr_size = stem_document_matrix->cols();

    if(pr_size > dvs.size())
        pr_size = dvs.size();

    float value = 0;

    for(int i=0; i< pr_size; i++)
    {
        int size = dvs[i]->stems().size();
        float tf = 0;
        for(int j=0;j<dvs[i]->stems().size();j++)
        {
            if(stem.compare(dvs[i]->stems()[j])==0) //Equal
            {
                tf += 1.0;
            }
            
            /*
              if(std::find(dvs[i]->stems().begin(), dvs[i]->stems().end(), stem)!=dvs[i]->stems().end())
            {
                has_stem += 1.0;
                }
            */       
        }
        delete dvs[i];

        value += tf / size;
    }    

    return value;
}

float FeatureGenerator::compute_relevance_weight(std::string &stem, float lambda)
{
    std::vector<indri::api::DocumentVector*> dvs = env.documentVectors(res_docids);
    int pr_size = stem_document_matrix->cols();

    if(pr_size > dvs.size())
        pr_size = dvs.size();

    float value = 0;

    for(int i=0; i< pr_size; i++)
    {
        int size = dvs[i]->stems().size();
        float tf = 0;
        for(int j=0;j<dvs[i]->stems().size();j++)
        {
            if(stem.compare(dvs[i]->stems()[j])==0) //Equal
            {
                tf += 1.0;
            }
            
        }
        delete dvs[i];

        value += tf * exp(res_scores[i]) / size;
    }    

    return value;
}

float FeatureGenerator::compute_df(std::string &stem)
{
    int df = env.documentStemCount(stem);
    int N = env.documentCount();

    return 1.0 * df / N;
}

float FeatureGenerator::compute_expr_df(std::string& expr)
{
    int df = env.documentExpressionCount( expr );
    int N = env.documentCount();
    return 1.0 * df / N;
}

float FeatureGenerator::compute_idf(std::string &stem)
{
    // Le's imple.?  To get his original idf, use this, but maybe bugged!!!
    //int df = env.documentCount(stem);
    
    //Mine
    int df = env.documentStemCount(stem);
    int N = env.documentCount();
    return Utils::IDF(df, N);
}

float FeatureGenerator::compute_expr_idf(std::string& expr)
{
    int df = env.documentExpressionCount( expr );
    int N = env.documentCount();
    return Utils::IDF(df, N);
}

float FeatureGenerator::compute_term_centrality(std::string &stem, const Eigen::MatrixXf& scm)
{
    //std::cout<<stem<<":"<<concept_distance(stem, stem,scm)<<std::endl;
    
    // Le's
    return concept_distance(stem, 1, scm);
    // Mine
    float stem_norm =concept_distance(stem, stem, scm);
    return concept_distance(stem, 1, scm) / sqrt(stem_norm);
    //return sqrt(concept_distance(stem, stem, scm));
}

float FeatureGenerator::compute_qterm_centrality(std::string& stem, const Eigen::MatrixXf& scm)
{
    float least_possible_value = 0;
    if(local_stem_dict.find(stem) == local_stem_dict.end())
    {
        return least_possible_value;
    }

    // *Terms*  should be stored in output_terms
    // *Stemmed* exprs should be stored in output_phrases

    int tid1 = local_stem_dict[stem];
    
    float max_dist = FLT_MIN;
    for(int i=0;i<output_terms.size();i++)
    {
        std::string query_stem = stem_it(output_terms[i]);// Should be stemmed first
        if(query_stem.compare(stem)!=0 && local_stem_dict.find(query_stem) != local_stem_dict.end())
        {
            int tid2 = local_stem_dict[query_stem];
            float dist = concept_distance_by_tid(tid1, tid2, scm);
            if(dist > max_dist)
                max_dist = dist;
        }
    }


    for(std::set<std::string>::iterator it=output_phrase_set.begin(); it!=output_phrase_set.end();it++)
    {
        std::string stemmed_expr = *it;
        if(stem.compare(stemmed_expr)!=0 && local_stem_dict.find(stemmed_expr) != local_stem_dict.end())
        {
            int tid2 = local_stem_dict[stemmed_expr];
            float dist = concept_distance_by_tid(tid1, tid2, scm);
            if(dist > max_dist)
                max_dist = dist;
        }
    }


    return max_dist;
}

float FeatureGenerator::compute_concept_centrality(std::string &stem, const Eigen::MatrixXf& scm)
{
    // c is fixed to 5 from pilot study
    int c = 5;
    
    // Le set it to 4 because of bug???
    c = 4;
    float v = 0;
    //    float divisor = sqrt(concept_distance(stem, stem, scm)) / c;
    //float stem_norm =concept_distance(stem, stem, scm);
    //float divisor = sqrt(stem_norm) * c;
    for(int i=2; i<= c + 1;i++)
    {
        std::string neighbor = concept_neighbor(stem, i, scm);
        //Le's
        v += concept_distance(stem, i, scm);

        //std::cout<<"Distance of "<<stem<<" "<<neighbor<<":"<<concept_distance(stem, i, scm)<<std::endl;
        //Mine
        // v += concept_distance(stem, i, scm) / divisor;
    }

    //Le's
    return v / c;
    //Mine
    return v;
}

float FeatureGenerator::compute_qconcept_centrality(std::string& stem, const Eigen::MatrixXf& scm)
{

    // c is fixed to 5 from pilot study
    int c = 5;
    
    // Le set it to 4 because of bug???
    //c = 3;

    int output_size = output_phrase_set.size() + output_stem_set.size();
    if (c>output_size)
        c = output_size;
    
    float v = 0;
    //    float divisor = sqrt(concept_distance(stem, stem, scm)) / c;
    //float stem_norm =concept_distance(stem, stem, scm);
    //float divisor = sqrt(stem_norm) * c;
    for(int i=2; i<= c + 1;i++)
    {
        std::string neighbor = concept_query_neighbor(stem, i, scm);
        //Le's
        v += concept_distance(stem, i, scm);

        //std::cout<<"Distance of "<<stem<<" "<<neighbor<<":"<<concept_distance(stem, i, scm)<<std::endl;
        //Mine
        // v += concept_distance(stem, i, scm) / divisor;
    }

    //Le's
    return v / c;
    //Mine
    return v;

}


float FeatureGenerator::compute_replaceability(std::string &stem, const Eigen::MatrixXf& scm)
{
    int c = 5;
    c = 3;
    float v = 0;
    float divisor = concept_distance(stem, stem, scm);

    for(int i=1; i<= c +1;i++)
    {
        std::string neighbor = concept_neighbor(stem, i, scm);

        // Neighbor term is just the stem itself
        if(neighbor.compare("[OOV]") == 0 || stem.compare(neighbor)==0)
        {
            continue;
        }


        int df = -1;
        //int df = env.documentCount(neighbor);
        if (neighbor[0]!='#') // Neighbor is a term
            df = env.documentStemCount(neighbor);
        else // Neighbor is an expression
             // Need to find its raw form to call documentExpressionCount
             // Seems that we can only iterate the expressionl list
        {

            for(int j=0;j<output_phrases_unique.size();j++)
            {
                std::string expr = output_phrases_unique[j];
                std::string stemmed_expr = stem_expr(expr);
                if(neighbor.compare(stemmed_expr) == 0)
                {
                    df = env.documentExpressionCount(expr);
                    break;
                }
            }
            std::cout<<"Neighbor of "<<stem<<" is "<<neighbor<<" DF:"<<df<<std::endl;
        }

        if(df == 0)// Wierd, why would some top term has 0 df? Answer: because of stemming
        {
            // My imple.
            /* 
            std::cout<<neighbor<<std::endl;
            continue;
            */

            // Le's imple.
            df = 1;
        }

        // count how many documents contains both stems
        //Mine
        //std::string band_query = "#band(\"" + stem + "\" \"" + neighbor + "\")";
        //Le's
        std::string band_query = "#band(" + stem + " " + neighbor + ")";
        float band_df = env.documentExpressionCount(band_query);
        std::cout<<stem<<" "<<neighbor<<" DF of "<<band_query<<" is:"<<int(band_df)<<std::endl;

        // Le's
        v += ((float)df - band_df) * concept_distance(stem, neighbor, scm) / df / divisor;

        //Mine
        //v += ((float)df - band_df) * concept_distance(stem, neighbor, scm) / df / divisor;
                
    }

    return v;
}

float FeatureGenerator::compute_qreplaceability(std::string& stem, const Eigen::MatrixXf& scm)
{
    int c = 5;

    int output_size = output_phrase_set.size() + output_stem_set.size();
    if (c>output_size)
        c = output_size;

    float v = 0;
    float divisor = concept_distance(stem, stem, scm);

    for(int i=1; i<= c +1;i++)
    {
        std::string neighbor = concept_query_neighbor(stem, i, scm);

        // Neighbor term is just the stem itself
        if(neighbor.compare("[OOV]") == 0 || stem.compare(neighbor)==0)
        {
            continue;
        }

        int df = -1;
        //int df = env.documentCount(neighbor);
        if (neighbor[0]!='#') // Neighbor is a term
            df = env.documentStemCount(neighbor);
        else // Neighbor is an expression
             // Need to find its raw form to call documentExpressionCount
             // Seems that we can only iterate the expressionl list
        {
            for(int j=0;j<output_phrases_unique.size();j++)
            {
                std::string expr = output_phrases_unique[j];
                std::string stemmed_expr = stem_expr(expr);
                if(neighbor.compare(stemmed_expr) == 0)
                {
                    df = env.documentExpressionCount(expr);
                    break;
                }
            }
            std::cout<<"Neighbor of "<<stem<<" is "<<neighbor<<" DF:"<<df<<std::endl;
        }

        if(df == 0)// Wierd, why would some top term has 0 df? Answer: because of stemming
        {
            // My imple.
            /* 
            std::cout<<neighbor<<std::endl;
            continue;
            */

            // Le's imple.
            df = 1;
        }

        // count how many documents contains both stems
        //Mine
        //std::string band_query = "#band(\"" + stem + "\" \"" + neighbor + "\")";
        //Le's
        std::string band_query = "#band(" + stem + " " + neighbor + ")";
        float band_df = env.documentExpressionCount(band_query);
        std::cout<<stem<<" "<<neighbor<<" DF of "<<band_query<<" is:"<<int(band_df)<<std::endl;

        // Le's
        v += ((float)df - band_df) * concept_distance(stem, neighbor, scm) / df / divisor;

        //Mine
        //v += ((float)df - band_df) * concept_distance(stem, neighbor, scm) / df / divisor;
                
    }

    return v;
}


void FeatureGenerator::load_plsa_result(std::string &wt_file, std::string &stem_file)
{
    std::ifstream wt_in(wt_file.data(), std::ifstream::in);
    /*
    std::ifstream stem_dict_in(stem_file.data(), std::ifstream::in);
    
    string line;

    while(getline(stem_dict_in, line))
    {
        std::istringstream iss(line);
        string stem;
        int id;
        iss>>stem>>id;
        local_stem_dict[stem] = id - 1;
    }

    stem_dict_in.close();
    */
    string line;
    std::vector<Eigen::Triplet<float> > tripleList;    
    int row = 0;
    int col = 0;
    while(getline(wt_in, line))
    {
        std::istringstream iss(line);
        col = 0;
        double value ;
        while(iss>>value)
        {
            tripleList.push_back(Eigen::Triplet<float>(row, col, value));
            col++;
        }
        row++;
    }
    wt_in.close();

    Eigen::SparseMatrix<float> plsa_wt_matrix(row, col);
    plsa_wt_matrix.setFromTriplets(tripleList.begin(), tripleList.end());
    
    stem_concept_plsa_matrix = plsa_wt_matrix;
}

void FeatureGenerator::load_dep(std::string &dep_file)
{
    std::ifstream dep_in(dep_file.c_str(), std::ifstream::in);
    string line;
    while(getline(dep_in, line))
    {
        if(line.length() == 0)
        {
            continue;
        }
        
        std::istringstream iss(line);
        int start = line.find_first_of('(') + 1;
        int sub_length = line.find_first_of('-') - start;
        std::string source = line.substr(start, sub_length);
        
        start = line.find_first_of(' ') + 1;
        sub_length = line.find_last_of('-') - start;
        std::string target = line.substr(start, sub_length);
        query_stem_dp_source.insert(stem_it(source));
        query_stem_dp_all.insert(stem_it(source));
        query_stem_dp_all.insert(stem_it(target));
        query_terms.insert(source);
        query_terms.insert(target);
    }

    dep_in.close();
}
 
float FeatureGenerator::compute_abstractness(std::string &stem)
{
    // The stem is among the target nodes but not among the source nodes -> leaf node
    if(query_stem_dp_source.find(stem) == query_stem_dp_source.end() && query_stem_dp_all.find(stem) != query_stem_dp_all.end() )
    {
        return 1.0;
    }

    // Other stems appearing in the target nodes or source nodes -> non-leaf nodes
    if(query_stem_dp_all.find(stem) != query_stem_dp_all.end() )
    {
        return 0.0;
    }

    // The stem is not in the output of the dep tree -> no abstractness
    // This is used to check whether this term would appear in features or not
    // If equals to -1, do not include this term in the features
    return -1.0;
}

float FeatureGenerator::compute_qtf(std::string &stem)
{
    std::set<std::string>::iterator it;
    float tf = 0;
    for (it = query_terms.begin(); it !=query_terms.end();it++)
    {
        std::string term = (*it);
        std::string stm = stem_it(term);

        if(stm.compare(stem)==0)
            tf += 1.0;
    }

    return tf / query_terms.size();
}

/*
std::string FeatureGenerator::generate_svm_features_le()
{
    std::ostringstream oss;
    std::vector<std::string>::iterator it;
    
    //    for (it=query_terms.begin() ; it != query_terms.end(); it++ )
    for (it=output_terms.begin() ; it != output_terms.end(); it++ )
    {
        std::string term = (*it);
        std::string stem = stem_it(term);

        // Check if stop word
        std::string termbkp(term);
        std::transform(termbkp.begin(), termbkp.end(), termbkp.begin(), ::tolower);
        if(stop_words.find(termbkp) != stop_words.end())
            continue;
        
        // If query stem does not appear in top documents,
        // ignore it
        if(local_stem_dict.find(stem) != local_stem_dict.end())
        {
            float rc = compute_true_recall(stem);
            rc = rc / (1-rc);//???
            float idf = compute_idf(stem);
            idf = idf / 10 - 0.5;//???
            float tcentra = compute_term_centrality(stem, stem_concept_matrix);
            tcentra = tcentra - 0.5;//???
            float ccentra = compute_concept_centrality(stem, stem_concept_matrix);
            ccentra = ccentra * 5 - 0.5;//???
            float rep = compute_replaceability(stem, stem_concept_matrix);
            rep = rep - 0.5;//???
            float abs = compute_abstractness(stem);
            abs = abs  - 0.5;//???

            //float qtf = compute_qtf(stem);
            float prf_df = compute_pseudo_recall(stem);
            float prf_ctf = compute_pseudo_recall_tf(stem);

            
            
            
            //float tcentra = compute_term_centrality(stem, stem_concept_plsa_matrix);
            //float ccentra = compute_concept_centrality(stem, stem_concept_plsa_matrix);
            //float rep = compute_replaceability(stem, stem_concept_plsa_matrix);

            //            float ptcentra = compute_term_centrality(stem, stem_concept_plsa_matrix);
            //            float pccentra = compute_concept_centrality(stem, stem_concept_plsa_matrix);
            //            float prep = compute_replaceability(stem, stem_concept_plsa_matrix);
                
            
            oss<<rc<<" 1:"<<idf<<" 2:"<<tcentra<<" 3:"<<ccentra<<" 4:"<<rep<<" 5:"<<abs<<" 69:-0.5";//<<" 6:"<<prf_df<<" 7:"<<prf_ctf;
            oss<<" #"<<stem<<" "<<term<<" "<<queryid<<endl;
            //oss<<rc<<" 1:"<<idf<<" 2:"<<tcentra<<" 3:"<<ccentra<<" 4:"<<rep<<" 5:"<<ptcentra<<" 6:"<<pccentra<<" 7:"<<prep<<" 8:"<<abs<<" #"<<stem<<" "<<term<<" "<<queryid<<endl;
            
        }
    }

    return oss.str();
}
*/

std::string FeatureGenerator::generate_svm_features()
{
    std::ostringstream oss;
    std::set<std::string>::iterator it;
    
    for (it=query_terms.begin() ; it != query_terms.end(); it++ )
        //for (it=output_terms.begin() ; it != output_terms.end(); it++ )
    {
        std::string term = (*it);
        std::string stem = stem_it(term);

        // Check if stop word
        std::string termbkp(term);
        std::transform(termbkp.begin(), termbkp.end(), termbkp.begin(), ::tolower);
        if(stop_words.find(termbkp) != stop_words.end())
            continue;
        
        // If query stem does not appear in top documents,
        // ignore it
        if(local_stem_dict.find(stem) != local_stem_dict.end())
        {
            // Use le's computed true recall
            //float rc = output_recalls[term];

            // This is mine
            float rc = compute_true_recall(stem);
            rc = rc / (1-rc);//???

            
            float idf = compute_idf(stem);
            idf = idf / 10 - 0.5;//???
            float tcentra = compute_term_centrality(stem, stem_concept_matrix);
            tcentra = tcentra - 0.5;//???
            float ccentra = compute_concept_centrality(stem, stem_concept_matrix);
            ccentra = ccentra * 5 - 0.5;//???
            float rep = compute_replaceability(stem, stem_concept_matrix);
            rep = rep - 0.5;//???
            float abs = compute_abstractness(stem);
            abs = abs  - 0.5;//???

            //float qtf = compute_qtf(stem);
            //float prf_df = compute_pseudo_recall(stem);
            //float prf_ctf = compute_pseudo_recall_tf(stem);
            
            //float tcentra = compute_term_centrality(stem, stem_concept_plsa_matrix);
            //float ccentra = compute_concept_centrality(stem, stem_concept_plsa_matrix);
            //float rep = compute_replaceability(stem, stem_concept_plsa_matrix);

            //            float ptcentra = compute_term_centrality(stem, stem_concept_plsa_matrix);
            //            float pccentra = compute_concept_centrality(stem, stem_concept_plsa_matrix);
            //            float prep = compute_replaceability(stem, stem_concept_plsa_matrix);
                
            oss<<rc<<" 1:"<<idf<<" 2:"<<tcentra<<" 3:"<<ccentra<<" 4:"<<rep<<" 5:"<<abs<<" 69:-0.5";//???
            //oss<<" 81:"<<prf_df<<" 82:"<<prf_ctf;
            //oss<<rc<<" 1:"<<idf<<" 2:"<<tcentra<<" 3:"<<ccentra<<" 4:"<<rep<<" 5:"<<abs<<" 69:0";//<<" 6:"<<prf_df<<" 7:"<<prf_ctf;
            oss<<" #"<<stem<<"\t"<<term<<"\t"<<queryid<<endl;
            //oss<<rc<<" 1:"<<idf<<" 2:"<<tcentra<<" 3:"<<ccentra<<" 4:"<<rep<<" 5:"<<ptcentra<<" 6:"<<pccentra<<" 7:"<<prep<<" 8:"<<abs<<" #"<<stem<<" "<<term<<" "<<queryid<<endl;
            
        }
    }

    return oss.str();
}

void FeatureGenerator::generate_true_recalls(std::string& true_recall_file)
{
    std::ofstream ofs(true_recall_file.data(), std::ofstream::out);
    std::vector<std::string>::iterator it;
    
    // For terms
    for (it=output_terms.begin() ; it != output_terms.end(); it++ )
    {
        std::string term = (*it);
        std::string stem = stem_it(term);
        
        float rc = compute_true_recall(stem);
        rc = rc / (1-rc);//???

        ofs<<rc<<" #"<<stem<<"\t"<<term<<"\t"<<queryid<<endl;

    }

    // For phrases
    for(it=output_phrases.begin(); it!= output_phrases.end();it++)
    {
        std::string phrase = (*it);
        std::string stemmed_expr = stem_expr(phrase);

        float rc = compute_expr_true_recall(phrase);
        rc = rc / (1-rc);//???
        
        ofs<<rc<<" #"<<stemmed_expr<<"\t"<<phrase<<"\t"<<queryid<<endl;
    }

    ofs.close();
}

std::string FeatureGenerator::generate_svm_features_with_phrase()
{
    std::ostringstream oss;
    std::vector<std::string>::iterator it;
    std::cout<<"PRF LENGTH:\t"<<prf_total_length<<std::endl;    
    // For individual terms
    //    for (it=query_terms.begin() ; it != query_terms.end(); it++ )
    for (it=output_terms.begin() ; it != output_terms.end(); it++ )
    {
        std::string term = (*it);
        std::string stem = stem_it(term);

        // Check if stop word
        std::string termbkp(term);
        std::transform(termbkp.begin(), termbkp.end(), termbkp.begin(), ::tolower);
        if(stop_words.find(termbkp) != stop_words.end())
            continue;
        
        // If query stem does not appear in top documents,
        // ignore it
        if(local_stem_dict.find(stem) != local_stem_dict.end())
        {
            // Output prf tf
            std::cout<<"PRF TF:\t"<<stem<<"\t"<<expr_prf_tf_map[stem]<<std::endl;
            
            // Use le's computed true recall
            //float rc = output_recalls[term];

            // This is mine
            float rc = compute_true_recall(stem);
            rc = rc / (1-rc);//???

            float rdf = compute_df(stem);
            
            float idf = compute_idf(stem);
            //idf = idf / 10 - 0.5;//SIGIR COMMENT
            float tcentra = compute_term_centrality(stem, stem_concept_matrix);
            //tcentra = tcentra - 0.5;//SIGIR COMMENT
            float ccentra = compute_concept_centrality(stem, stem_concept_matrix);
            //ccentra = ccentra * 5 - 0.5;//SIGIR COMMENT
            float rep = compute_replaceability(stem, stem_concept_matrix);
            //rep = rep - 0.5;//SIGIR COMMENT
                
            oss<<rc<<" 1:"<<idf<<" 2:"<<tcentra<<" 3:"<<ccentra<<" 4:"<<rep<<" 69:-0.5";//???
            oss<<" 101:"<<rdf<<" 102:"<<(1-rdf)<<" 103:"<<(1-rdf)*(1-rdf)/rdf<<" 104:"<<(1-rdf)*(1-rdf)*(1-rdf)/rdf/rdf;

            float qtcentra = compute_qterm_centrality(stem, stem_concept_matrix);
            float qccentra = compute_qconcept_centrality(stem, stem_concept_matrix);
            float qrep = compute_qreplaceability(stem, stem_concept_matrix);
            
            oss<<" 201:"<<qtcentra<<" 202:"<<qccentra<<" 203:"<<qrep;
            
            //oss<<" 81:"<<prf_df<<" 82:"<<prf_ctf;
            //oss<<rc<<" 1:"<<idf<<" 2:"<<tcentra<<" 3:"<<ccentra<<" 4:"<<rep<<" 5:"<<abs<<" 69:0";//<<" 6:"<<prf_df<<" 7:"<<prf_ctf;
            oss<<" #"<<stem<<"\t"<<term<<"\t"<<queryid<<endl;
            //oss<<rc<<" 1:"<<idf<<" 2:"<<tcentra<<" 3:"<<ccentra<<" 4:"<<rep<<" 5:"<<ptcentra<<" 6:"<<pccentra<<" 7:"<<prep<<" 8:"<<abs<<" #"<<stem<<" "<<term<<" "<<queryid<<endl;
            
        }
    }

    // For phrase
    for(it=output_phrases.begin(); it!= output_phrases.end();it++)
    {
        std::string phrase = (*it);
        std::string stemmed_expr = stem_expr(phrase);

        int df = env.documentExpressionCount( phrase );
        
        // expr not exist in the dict/collection
        if(local_stem_dict.find(stemmed_expr ) == local_stem_dict.end())
        {
            continue;
        }

        // Output prf tf
        std::cout<<"PRF TF:\t"<<phrase<<"\t"<<expr_prf_tf_map[stemmed_expr]<<std::endl;

        int stemid = local_stem_dict[stemmed_expr];

        float rc = compute_expr_true_recall(phrase);
        rc = rc / (1-rc);//???

        float rdf = compute_expr_df(phrase);
        
        float idf = compute_expr_idf(phrase);
        //idf = idf / 10 -0.5;//SIGIR COMMENT
        float tcentra = compute_term_centrality(stemmed_expr, stem_concept_matrix);
        //tcentra = tcentra - 0.5;//SIGIR COMMENT
        float ccentra = compute_concept_centrality(stemmed_expr, stem_concept_matrix);
        //ccentra = ccentra * 5 - 0.5;//SIGIR COMMENT
        float rep = compute_replaceability(stemmed_expr, stem_concept_matrix);
        //rep = rep - 0.5;//SIGIR COMMENT
        
        oss<<rc<<" 1:"<<idf<<" 2:"<<tcentra<<" 3:"<<ccentra<<" 4:"<<rep<<" 69:-0.5";//???
        oss<<" 101:"<<rdf<<" 102:"<<(1-rdf)<<" 103:"<<(1-rdf)*(1-rdf)/rdf<<" 104:"<<(1-rdf)*(1-rdf)*(1-rdf)/rdf/rdf;

        float qtcentra = compute_qterm_centrality(stemmed_expr, stem_concept_matrix);
        float qccentra = compute_qconcept_centrality(stemmed_expr, stem_concept_matrix);
        float qrep = compute_qreplaceability(stemmed_expr, stem_concept_matrix);
        
        oss<<" 201:"<<qtcentra<<" 202:"<<qccentra<<" 203:"<<qrep;

        oss<<" #"<<stemmed_expr<<"\t"<<phrase<<"\t"<<queryid<<endl;
    }

    return oss.str();
}


std::string FeatureGenerator::generate_regression_features_with_phrase()
{
    std::ostringstream oss;
    std::vector<std::string>::iterator it;
    std::cout<<"PRF LENGTH:\t"<<prf_total_length<<std::endl;    

    int ctf_max = 0;
    int ctf_sum = 0;

    int df_max = 0;
    int df_sum = 0;

    // For individual terms
    //    for (it=query_terms.begin() ; it != query_terms.end(); it++ )
    for (it=output_terms.begin() ; it != output_terms.end(); it++ )
    {
        std::string term = (*it);
        std::string stem = stem_it(term);

        // Check if stop word
        std::string termbkp(term);
        std::transform(termbkp.begin(), termbkp.end(), termbkp.begin(), ::tolower);
        if(stop_words.find(termbkp) != stop_words.end())
            continue;

        // expr not exist in the dict/collection
        if(local_stem_dict.find(stem ) == local_stem_dict.end())
        {
            continue;
        }

        int ctf = env.stemCount(stem);
        if(ctf > ctf_max)
            ctf_max = ctf;
        ctf_sum += ctf;

        int df = env.documentStemCount(stem);

        if(df > df_max)
            df_max = df;
        df_sum += df;
    }

    // For phrase
    for(it=output_phrases.begin(); it!= output_phrases.end();it++)
    {
        std::string phrase = (*it);
        std::string stemmed_expr = stem_expr(phrase);

        // expr not exist in the dict/collection
        if(local_stem_dict.find(stemmed_expr ) == local_stem_dict.end())
        {
            continue;
        }

        int ctf = env.expressionCount( phrase );

        if(ctf > ctf_max)
            ctf_max = ctf;
        ctf_sum += ctf;

        int df = env.documentExpressionCount( phrase );
        if(df > df_max)
            df_max = df;
        df_sum += df;
    }

    int CNT = env.termCount();
        
    for (it=output_terms.begin() ; it != output_terms.end(); it++ )
    {
        std::string term = (*it);
        std::string stem = stem_it(term);

        // Check if stop word
        std::string termbkp(term);
        std::transform(termbkp.begin(), termbkp.end(), termbkp.begin(), ::tolower);
        if(stop_words.find(termbkp) != stop_words.end())
            continue;

        // expr not exist in the dict/collection
        if(local_stem_dict.find(stem ) == local_stem_dict.end())
        {
            continue;
        }

        float rc = compute_true_recall(stem);
        rc = rc / (1-rc);
        
        float idf = compute_idf(stem);
        float rdf = compute_df(stem);
        int ctf = env.stemCount(stem);
        int df = env.documentStemCount(stem);

        oss<<rc<<" 301:"<<idf<<" 302:"<<rdf;
        oss<<" 401:"<<ctf<<" 402:"<<(1.0*ctf/ctf_max)<<" 403:"<<(1.0*ctf/ctf_sum)<<" 404:"<<log(1.0*ctf)<<" 405:"<<log(1.0*ctf/ctf_max)<<" 406:"<<log(1.0*ctf/ctf_sum);
        oss<<" 501:"<<df<< " 502:"<<(1.0* df/ df_max)<<" 503:"<<(1.0* df/ df_sum)<<" 504:"<<log(1.0* df)<<" 505:"<<log(1.0* df/ df_max)<<" 506:"<<log(1.0* df/ df_sum);
        // residual idf
        int N = env.documentCount();
        oss<<" 601:"<<(log(1.0 * N / df) - log(1.0 /(1-exp(-1.0*ctf/N))));
       
        // compute wig
        std::vector<std::string> rules;
        rules.push_back("method:d,mu:0");
        env.setScoringRules(rules);

        float wig = 0;
        int wig_M = 50;
        std::vector< indri::api::ScoredExtentResult > concept_res = env.runQuery(term, wig_M);
        for(int j=0;j<concept_res.size();j++)
        {
            float score = concept_res[j].score;
            wig += score;
        }
        wig /= wig_M;
        wig -= log( 1.0 * ctf / CNT);
        wig /= (-log( 1.0 * ctf / CNT));
        oss<<" 701:"<<wig;
        oss<<" #"<<stem<<"\t"<<term<<"\t"<<queryid<<endl;
    }

    // For phrase
    for(it=output_phrases.begin(); it!= output_phrases.end();it++)
    {
        std::string phrase = (*it);
        std::string stemmed_expr = stem_expr(phrase);

        // expr not exist in the dict/collection
        if(local_stem_dict.find(stemmed_expr ) == local_stem_dict.end())
        {
            continue;
        }

        float rc = compute_expr_true_recall(phrase);
        rc = rc / (1-rc);

        float idf = compute_expr_idf( phrase );
        float rdf = compute_expr_df( phrase );
        int ctf = env.expressionCount( phrase );
        int df = env.documentExpressionCount( phrase );

        oss<<rc<<" 301:"<<idf<<" 302:"<<rdf;
        oss<<" 401:"<<ctf<<" 402:"<<(1.0*ctf/ctf_max)<<" 403:"<<(1.0*ctf/ctf_sum)<<" 404:"<<log(1.0*ctf)<<" 405:"<<log(1.0*ctf/ctf_max)<<" 406:"<<log(1.0*ctf/ctf_sum);
        oss<<" 501:"<<df<< " 502:"<<(1.0* df/ df_max)<<" 503:"<<(1.0* df/ df_sum)<<" 504:"<<log(1.0* df)<<" 505:"<<log(1.0* df/ df_max)<<" 506:"<<log(1.0* df/ df_sum);
        // residual idf
        int N = env.documentCount();
        oss<<" 601:"<<(log(1.0 * N/ df) - log(1.0 /(1-exp(-1.0*ctf/N))));
        
        // wig
        float wig = 0;
        int wig_M = 50;
        std::vector< indri::api::ScoredExtentResult > concept_res = env.runQuery(phrase, wig_M);
        for(int j=0;j<concept_res.size();j++)
        {
            float score = concept_res[j].score;
            wig += score;
        }
        wig /= wig_M;
        wig -= log( 1.0 * ctf / CNT);
        wig /= (-log( 1.0 * ctf / CNT));
        oss<<" 701:"<<wig;
        oss<<" #"<<stemmed_expr<<"\t"<<phrase<<"\t"<<queryid<<endl;
    }

    return oss.str();
}


std::string FeatureGenerator::generate_svm_features_SIGIR2014()
{
    std::ostringstream oss;
    std::vector<std::string>::iterator it;
    std::cout<<"PRF LENGTH:\t"<<prf_total_length<<std::endl;    

    int ctf_max = 0;
    int ctf_sum = 0;

    int df_max = 0;
    int df_sum = 0;

    // Get sums and maxes
    // For individual terms                                        
    for (it=output_terms.begin() ; it != output_terms.end(); it++ )
    {
        std::string term = (*it);
        std::string stem = stem_it(term);

        // Check if stop word                                                                                                        
        std::string termbkp(term);
        std::transform(termbkp.begin(), termbkp.end(), termbkp.begin(), ::tolower);
        if(stop_words.find(termbkp) != stop_words.end())
            continue;

        // expr not exist in the dict/collection                                                                                     
        if(local_stem_dict.find(stem ) == local_stem_dict.end())
        {
            continue;
        }

        int ctf = env.stemCount(stem);
        if(ctf > ctf_max)
            ctf_max = ctf;
        ctf_sum += ctf;

        int df = env.documentStemCount(stem);

        if(df > df_max)
            df_max = df;
        df_sum += df;
    }

    // For phrase                                                                                                                   
    for(it=output_phrases.begin(); it!= output_phrases.end();it++)
    {
        std::string phrase = (*it);
        std::string stemmed_expr = stem_expr(phrase);

        // expr not exist in the dict/collection                                                                                 
        if(local_stem_dict.find(stemmed_expr ) == local_stem_dict.end())
        {
            continue;
        }

        int ctf = env.expressionCount( phrase );

        if(ctf > ctf_max)
            ctf_max = ctf;
        ctf_sum += ctf;

        int df = env.documentExpressionCount( phrase );
        if(df > df_max)
            df_max = df;
        df_sum += df;
    }


    INT64 CNT = env.termCount();
        
    // Output features
    // For individual terms
    for (it=output_terms.begin() ; it != output_terms.end(); it++ )
    {
        std::string term = (*it);
        std::string stem = stem_it(term);

        // Check if stop word
        std::string termbkp(term);
        std::transform(termbkp.begin(), termbkp.end(), termbkp.begin(), ::tolower);
        if(stop_words.find(termbkp) != stop_words.end())
            continue;
        
        // If query stem does not appear in top documents,
        // ignore it
        if(local_stem_dict.find(stem) != local_stem_dict.end())
        {
            // Output prf tf
            std::cout<<"PRF TF:\t"<<stem<<"\t"<<expr_prf_tf_map[stem]<<std::endl;
            
            float rc = compute_true_recall(stem);
            rc = rc / (1-rc);

            float idf = compute_idf(stem);
            float rdf = compute_df(stem);
            int ctf = env.stemCount(stem);
            int df = env.documentStemCount(stem);

            float tcentra = compute_term_centrality(stem, stem_concept_matrix);
            float ccentra = compute_concept_centrality(stem, stem_concept_matrix);
            float rep = compute_replaceability(stem, stem_concept_matrix);

            // SVD features
            oss<<rc<<" 1:"<<idf<<" 2:"<<tcentra<<" 3:"<<ccentra<<" 4:"<<rep<<" 69:-0.5";

            // DF, 1-DF features
            oss<<" 101:"<<rdf<<" 102:"<<(1-rdf)<<" 103:"<<(1-rdf)*(1-rdf)/rdf<<" 104:"<<(1-rdf)*(1-rdf)*(1-rdf)/rdf/rdf;

            float qtcentra = compute_qterm_centrality(stem, stem_concept_matrix);
            float qccentra = compute_qconcept_centrality(stem, stem_concept_matrix);
            float qrep = compute_qreplaceability(stem, stem_concept_matrix);
            
            // Inq features
            oss<<" 201:"<<qtcentra<<" 202:"<<qccentra<<" 203:"<<qrep;

            // Regression Rank features
            oss<<" 401:"<<ctf<<" 402:"<<(1.0*ctf/ctf_max)<<" 403:"<<(1.0*ctf/ctf_sum)<<" 404:"<<log(1.0*ctf)<<" 405:"<<log(1.0*ctf/ctf_max)<<" 406:"<<log(1.0*ctf/ctf_sum);
            oss<<" 501:"<<df<< " 502:"<<(1.0* df/ df_max)<<" 503:"<<(1.0* df/ df_sum)<<" 504:"<<log(1.0* df)<<" 505:"<<log(1.0* df/ df_max)<<" 506:"<<log(1.0* df/ df_sum);

            // residual idf                                                                                                         
            int N = env.documentCount();
            oss<<" 601:"<<(log(1.0 * N / df) - log(1.0 /(1-exp(-1.0*ctf/N))));

            // compute wig                                                                                            
            std::vector<std::string> rules;
            rules.push_back("method:d,mu:0");
            env.setScoringRules(rules);

            float wig = 0;
            int wig_M = 100;
            std::vector< indri::api::ScoredExtentResult > concept_res = env.runQuery(term, wig_M);
            for(int j=0;j<concept_res.size();j++)
            {
                float score = concept_res[j].score;
                wig += score;
            }
            wig /= wig_M;
            float wig_tmp = log( 1.0 * ctf ) - log (1.0* CNT);
            cout<<"WIG_TMP: "<<wig_tmp<<"/t"<<log(1.0*ctf)<<"/t"<<log(1.0*CNT)<<endl;
            wig -= wig_tmp;
            wig /= (- wig_tmp);
            oss<<" 701:"<<wig;

            oss<<" #"<<stem<<"\t"<<term<<"\t"<<queryid<<endl;
        }
    }

    // For phrase
    for(it=output_phrases.begin(); it!= output_phrases.end();it++)
    {
        std::string phrase = (*it);
        std::string stemmed_expr = stem_expr(phrase);

        int df = env.documentExpressionCount( phrase );
        
        // expr not exist in the dict/collection
        if(local_stem_dict.find(stemmed_expr ) != local_stem_dict.end())
        {
            // Output prf tf
            std::cout<<"PRF TF:\t"<<phrase<<"\t"<<expr_prf_tf_map[stemmed_expr]<<std::endl;
            
            //            int stemid = local_stem_dict[stemmed_expr];
            
            float rc = compute_expr_true_recall(phrase);
            rc = rc / (1-rc);

            float idf = compute_expr_idf(phrase);            
            float rdf = compute_expr_df(phrase);
            int ctf = env.expressionCount( phrase );
            int df = env.documentExpressionCount( phrase );

            float tcentra = compute_term_centrality(stemmed_expr, stem_concept_matrix);
            float ccentra = compute_concept_centrality(stemmed_expr, stem_concept_matrix);
            float rep = compute_replaceability(stemmed_expr, stem_concept_matrix);

            // SVD features
            oss<<rc<<" 1:"<<idf<<" 2:"<<tcentra<<" 3:"<<ccentra<<" 4:"<<rep<<" 69:-0.5";

            // DF, 1-DF features
            oss<<" 101:"<<rdf<<" 102:"<<(1-rdf)<<" 103:"<<(1-rdf)*(1-rdf)/rdf<<" 104:"<<(1-rdf)*(1-rdf)*(1-rdf)/rdf/rdf;
            
            float qtcentra = compute_qterm_centrality(stemmed_expr, stem_concept_matrix);
            float qccentra = compute_qconcept_centrality(stemmed_expr, stem_concept_matrix);
            float qrep = compute_qreplaceability(stemmed_expr, stem_concept_matrix);
            
            // Inq features
            oss<<" 201:"<<qtcentra<<" 202:"<<qccentra<<" 203:"<<qrep;

            // Regression Rank features
            oss<<" 401:"<<ctf<<" 402:"<<(1.0*ctf/ctf_max)<<" 403:"<<(1.0*ctf/ctf_sum)<<" 404:"<<log(1.0*ctf)<<" 405:"<<log(1.0*ctf/ctf_max)<<" 406:"<<log(1.0*ctf/ctf_sum);
            oss<<" 501:"<<df<< " 502:"<<(1.0* df/ df_max)<<" 503:"<<(1.0* df/ df_sum)<<" 504:"<<log(1.0* df)<<" 505:"<<log(1.0* df/ df_max)<<" 506:"<<log(1.0* df/ df_sum);

            // Residual idf
            int N = env.documentCount();
            oss<<" 601:"<<(log(1.0 * N/ df) - log(1.0 /(1-exp(-1.0*ctf/N))));

            // wig                                                                                                                  
            float wig = 0;
            int wig_M = 100;
            std::vector< indri::api::ScoredExtentResult > concept_res = env.runQuery(phrase, wig_M);
            for(int j=0;j<concept_res.size();j++)
            {
                float score = concept_res[j].score;
                wig += score;
            }
            wig /= wig_M;
            float wig_tmp = log( 1.0 * ctf ) - log (1.0* CNT);
            cout<<"WIG_TMP: "<<wig_tmp<<"/t"<<log(1.0*ctf)<<"/t"<<log(1.0*CNT)<<endl;
            wig -= wig_tmp;
            wig /= (- wig_tmp);
            oss<<" 701:"<<wig;

            oss<<" #"<<stemmed_expr<<"\t"<<phrase<<"\t"<<queryid<<endl;
        }
    }

    return oss.str();
}

std::string FeatureGenerator::generate_svm_st_features()
{
    std::ostringstream oss;
    std::set<std::string>::iterator it;
    
    for (it=query_terms.begin() ; it != query_terms.end(); it++ )
    {
        std::string term = (*it);
        std::string stem = stem_it(term);

        // Check if stop word
        std::string termbkp(term);
        std::transform(termbkp.begin(), termbkp.end(), termbkp.begin(), ::tolower);
        if(stop_words.find(termbkp) != stop_words.end())
            continue;
        
        // If query stem does not appear in top documents,
        // ignore it
        if(local_stem_dict.find(stem) != local_stem_dict.end())
        {
            // Use le's computed true recall
            //float rc = output_recalls[term];

            // This is mine
            float rc = compute_true_recall(stem);
            //rc = rc / (1-rc);//???
            
            float idf = compute_idf(stem);
            //idf = idf / 10 - 0.5;//???
            float tcentra = compute_term_centrality(stem, stem_concept_matrix);
            //tcentra = tcentra - 0.5;//???
            float ccentra = compute_concept_centrality(stem, stem_concept_matrix);
            //ccentra = ccentra * 5 - 0.5;//???
            float rep = compute_replaceability(stem, stem_concept_matrix);
            //rep = rep - 0.5;//???
            float abs = compute_abstractness(stem);
            //abs = abs  - 0.5;//???

            //float qtf = compute_qtf(stem);
            //float prf_df = compute_pseudo_recall(stem);
            //float prf_ctf = compute_pseudo_recall_tf(stem);
            
            //float tcentra = compute_term_centrality(stem, stem_concept_plsa_matrix);
            //float ccentra = compute_concept_centrality(stem, stem_concept_plsa_matrix);
            //float rep = compute_replaceability(stem, stem_concept_plsa_matrix);

            //            float ptcentra = compute_term_centrality(stem, stem_concept_plsa_matrix);
            //            float pccentra = compute_concept_centrality(stem, stem_concept_plsa_matrix);
            //            float prep = compute_replaceability(stem, stem_concept_plsa_matrix);
                
            
            //oss<<rc<<" 1:"<<idf<<" 2:"<<tcentra<<" 3:"<<ccentra<<" 4:"<<rep<<" 5:"<<abs<<" 69:-0.5";
            oss<<rc<<" 1:"<<idf<<" 2:"<<tcentra<<" 3:"<<ccentra<<" 4:"<<rep<<" 5:"<<abs<<" 69:-1";
            //oss<<" 81:"<<prf_df<<" 82:"<<prf_ctf;
            //oss<<rc<<" 1:"<<idf<<" 2:"<<tcentra<<" 3:"<<ccentra<<" 4:"<<rep<<" 5:"<<abs<<" 69:0";//<<" 6:"<<prf_df<<" 7:"<<prf_ctf;
            oss<<" #"<<stem<<" "<<term<<" "<<queryid<<endl;
            //oss<<rc<<" 1:"<<idf<<" 2:"<<tcentra<<" 3:"<<ccentra<<" 4:"<<rep<<" 5:"<<ptcentra<<" 6:"<<pccentra<<" 7:"<<prep<<" 8:"<<abs<<" #"<<stem<<" "<<term<<" "<<queryid<<endl;
            
        }
    }

    return oss.str();
}


std::string FeatureGenerator::generate_svm_beta_features()
{
    std::ostringstream oss;
    std::vector<std::string>::iterator it;
    
    //    for (it=query_terms.begin() ; it != query_terms.end(); it++ )
    for (it=output_terms.begin() ; it != output_terms.end(); it++ )
    {
        std::string term = (*it);
        std::string stem = stem_it(term);

        // Check if stop word
        std::string termbkp(term);
        std::transform(termbkp.begin(), termbkp.end(), termbkp.begin(), ::tolower);
        if(stop_words.find(termbkp) != stop_words.end())
            continue;
        
        // If query stem does not appear in top documents,
        // ignore it
        if(local_stem_dict.find(stem) != local_stem_dict.end())
        {
            // Use le's computed true recall
            // float rc = output_recalls[term];

            // This is mine
            float rc = compute_true_recall(stem);
            rc = rc / (1-rc);//???
            
            float idf = compute_idf(stem);
            //idf = idf / 10 - 0.5;//???
            float tcentra = compute_term_centrality(stem, stem_concept_matrix);
            //tcentra = tcentra - 0.5;//???
            float ccentra = compute_concept_centrality(stem, stem_concept_matrix);
            //ccentra = ccentra * 5 - 0.5;//???
            float rep = compute_replaceability(stem, stem_concept_matrix);
            //rep = rep - 0.5;//???
            float abs = compute_abstractness(stem);
            //abs = abs  - 0.5;//???

            //float qtf = compute_qtf(stem);
            float prf_df = compute_pseudo_recall(stem);
            float prf_ctf = compute_pseudo_recall_tf(stem);

            // compute covered query term number features
            int stem_id = local_stem_dict[stem];
            float cqtn_avg = 0;
            float cqtn_min = FLT_MAX;
            float cqtn_max = -1;

            for(int i=0;i<stem_document_tf_matrix->cols();i++)
            {
                int tf = stem_document_tf_matrix->coeffRef(stem_id, i);
                cqtn_avg += tf;
                if(tf < cqtn_min)
                    cqtn_min = tf;
                if(tf > cqtn_max)
                    cqtn_max = tf;
            }
            cqtn_avg /= stem_document_tf_matrix->cols();

            // compute covered query term rative frequency features
            float cqtrf_avg = 0;
            float cqtrf_min = FLT_MAX;
            float cqtrf_max = -1;

            for(int i=0;i<stem_document_tf_matrix->cols();i++)
            {
                int tf = stem_document_tf_matrix->coeffRef(stem_id, i);
                float rtf = (float) tf / document_lens[i];
                cqtrf_avg += rtf;
                if(tf < cqtrf_min)
                    cqtrf_min = tf;
                if(tf > cqtrf_max)
                    cqtrf_max = tf;
            }
            cqtrf_avg /= stem_document_tf_matrix->cols();

            // compute tfidf features
            float tfidf_avg = 0;
            float tfidf_min = FLT_MAX;
            float tfidf_max = -1;
            
            for(int i=0;i<stem_document_matrix->cols();i++)
            {
                float tfidf = stem_document_matrix->coeffRef(stem_id, i);
                tfidf_avg += tfidf;
                if(tfidf < tfidf_min)
                    tfidf_min = tfidf;
                if(tfidf > tfidf_max)
                    tfidf_max = tfidf;
            }
            tfidf_avg /= stem_document_matrix->cols();
            
            //float ptcentra = compute_term_centrality(stem, stem_concept_plsa_matrix);
            //float pccentra = compute_concept_centrality(stem, stem_concept_plsa_matrix);
            //float prep = compute_replaceability(stem, stem_concept_plsa_matrix);
            
            oss<<rc<<" 1:"<<idf<<" 2:"<<tcentra<<" 3:"<<ccentra<<" 4:"<<rep<<" 5:"<<abs<<" 69:-0.5";
            oss<<" 81:"<<prf_df<<" 82:"<<prf_ctf;
            //oss<<" 91:"<<ptcentra<<" 92:"<<pccentra<<" 93:"<<prep;
            //oss<<rc<<" 1:"<<idf<<" 2:"<<tcentra<<" 3:"<<ccentra<<" 4:"<<rep<<" 5:"<<abs<<" 69:0";//<<" 6:"<<prf_df<<" 7:"<<prf_ctf;
            oss<<" 101:"<<cqtn_avg<<" 102:"<<cqtn_min<<" 103:"<<cqtn_max<<" 104:"<<cqtrf_avg<<" 105:"<<cqtrf_min<<" 106:"<<cqtrf_max<<" 107:"<<tfidf_avg<<" 108:"<<tfidf_min<<" 109:"<<tfidf_max;
            oss<<" #"<<stem<<" "<<term<<" "<<queryid<<endl;
            //oss<<rc<<" 1:"<<idf<<" 2:"<<tcentra<<" 3:"<<ccentra<<" 4:"<<rep<<" 5:"<<ptcentra<<" 6:"<<pccentra<<" 7:"<<prep<<" 8:"<<abs<<" #"<<stem<<" "<<term<<" "<<queryid<<endl;
        }
    }

    return oss.str();
}


std::string FeatureGenerator::generate_pr_recall_query()
{
    std::ostringstream oss;

    std::set<std::string>::iterator it;
    oss<<"\t<query><type>indri</type><number>"<<queryid<<"</number><text>"<<endl;
    oss<<"#weight( ";
    
    for (it=query_terms.begin() ; it != query_terms.end(); it++ )
    {
        std::string term = (*it);
        std::string stem = stem_it(term);

        // If query stem does not appear in top documents,
        // ignore it
        if(local_stem_dict.find(stem) != local_stem_dict.end())
        {
            //float prc = compute_weighted_pseudo_recall(stem);
            /* Here for recall tf*/
            //float prc = compute_pseudo_recall_tf(stem);
            //float prc = compute_pseudo_recall(stem);
            float prc = compute_weighted_pseudo_recall_tf(stem);
            //float prc = compute_rocchio_weight(stem, 0);
            //float prc = compute_relevance_weight(stem, 0);
            //float prc = compute_true_recall(stem);
            //float rc = compute_true_recall(stem);

            //float idf = compute_idf(stem);
            oss<<fixed<<setprecision(16)<<prc<<" "<<term<<" ";
            //oss<<prc<<" "<<term<<" ";

            //            oss<<prc<<" 1:"<<rc<<" 2:"<<idf<<" #"<<stem<<" "<<term<<" "<<queryid<<endl;
        }
    }

    oss<<" )"<<endl;
    oss<<"\t</text></query>"<<endl;

    return oss.str();
}

std::string FeatureGenerator::generate_rf_index_recall_query()
{
    std::ostringstream oss;

    std::set<std::string>::iterator it;
    oss<<"\t<query><type>indri</type><number>"<<queryid<<"</number><text>"<<endl;
    oss<<"#weight( ";

    std::vector<std::string> query_stems;
    get_query_stems(query_stems);
    
    for (it=query_terms.begin() ; it != query_terms.end(); it++ )
    {
        std::string term = (*it);
        std::string stem = stem_it(term);

        // If query stem does not appear in top documents,
        // ignore it
        if(local_stem_dict.find(stem) != local_stem_dict.end())
        {
            float irc = compute_true_index_recall(stem, query_stems);
            //            float rc = compute_true_recall(stem);

            //float idf = compute_idf(stem);

            oss<<irc<<" "<<term<<" ";

            //            oss<<prc<<" 1:"<<rc<<" 2:"<<idf<<" #"<<stem<<" "<<term<<" "<<queryid<<endl;
        }
    }

    oss<<" )"<<endl;
    oss<<"\t</text></query>"<<endl;

    return oss.str();
}

std::string FeatureGenerator::generate_pr_guided_recall_query(int n_rel, std::string &partial_true_file)
{
    std::ostringstream oss;
    std::ofstream partial_true_out(partial_true_file.data(), std::ofstream::out);

    std::set<std::string>::iterator it;
    oss<<"\t<query><type>indri</type><number>"<<queryid<<"</number><text>"<<endl;
    oss<<"#weight( ";

    partial_true_out<<"\t<query><type>indri</type><number>"<<queryid<<"</number><text>"<<endl;
    partial_true_out<<"#weight( ";

    
    // Generate guided information
    std::vector<std::string> rel_seed_docs;
    std::set<std::string> rel_seed_sets;
    std::vector<lemur::api::DOCID_T> rel_seed_docids;

    // Two way to construct relevant document seeds
    // 1. To use a small subset of relevant judgements

    // BEGIN METHOD 1
    /*
    
    if(n_rel > qrel_docs.size())
        n_rel = qrel_docs.size();

    srand(time(NULL));
    while(rel_seed_sets.size() < n_rel)
    {
        int index = rand() % qrel_docs.size();
        rel_seed_sets.insert(qrel_docs[index]);
    }

    for(it=rel_seed_sets.begin() ; it!=rel_seed_sets.end();it++)
    {
        rel_seed_docs.push_back(*it);
    }
    rel_seed_docids = env.documentIDsFromMetadata("docno", rel_seed_docs);
    */
    //END METHOD 1
    

    // 2. To assume documents containing all query terms are relevant
    // BEGIN METHOD 2

    std::vector<argsort_pair> pair_list;

    std::vector<indri::api::DocumentVector*> dvs = env.documentVectors(res_docids);    
    int pr_size = stem_document_matrix->cols();

    if(pr_size > dvs.size())
        pr_size = dvs.size();

    for(int i=0;i<pr_size;i++)
    {
        bool hasAllTerms = true;
        int num_of_terms_contained = 0;
        for(it=query_terms.begin();it!=query_terms.end();it++)
        {
            std::string term = (*it);
            std::string stem = stem_it(term);
            int stemid = local_stem_dict[stem];
            float tfidf = stem_document_matrix->coeffRef(stemid,i);
            if(tfidf != 0)
            {
                num_of_terms_contained ++;
            }
            if(tfidf == 0)
            {
                hasAllTerms = false;
                break;
            }
        }
        
        pair_list.push_back(argsort_pair(i, num_of_terms_contained));
    
       // if(hasAllTerms)
       // {
        //    if(rel_seed_docids.size() < n_rel)
          //      rel_seed_docids.push_back(res_docids[i]);
          //  else
            //    break;
       // }
    
    }

    std::sort(pair_list.begin(), pair_list.end(), argsort_comp);
    for(int i=0;i<n_rel && i<pair_list.size();i++)
    {
        rel_seed_docids.push_back(res_docids[pair_list[i].first]);
    }
    
    if(rel_seed_docids.size() ==0)
        std::cout<<"ERROR, NO REL!!!"<<std::endl;

    // END METHOD 2

    // Get relevant seeds tfidf matrix

    std::vector<Eigen::Triplet<float> > tripleList;
    tripleList.reserve(rel_seed_docids.size() * 500);
    
    std::vector<indri::api::DocumentVector*> rel_seed_dvs;
    rel_seed_dvs = env.documentVectors(rel_seed_docids);
    
    int stemid_index = local_stem_list.size();
    int N = env.documentCount();

    for(int i=0;i<rel_seed_dvs.size();i++)
    {
        int docLength = rel_seed_dvs[i]->positions().size();
        int unique_stem_count = rel_seed_dvs[i]->stems().size();

        std::vector<int> tfs(unique_stem_count, 0);
        for(int j=0;j<rel_seed_dvs[i]->positions().size();j++)
        {
            int pos = rel_seed_dvs[i]->positions()[j];
            tfs[pos]++;
        }

        for(int j=0;j<rel_seed_dvs[i]->stems().size();j++)
        {
            std::string stem = rel_seed_dvs[i]->stems()[j];
            int stemid = -1;
            if(local_stem_dict.find(stem) == local_stem_dict.end())
            {
                stemid = stemid_index;
                local_stem_list.push_back(stem);
                local_stem_dict[stem] = stemid_index++;
            }
            else
            {
                stemid = local_stem_dict[stem];
            }

            int tf = tfs[j];
            int df = env.documentStemCount(stem);
            float tfidf = Utils::TFIDF(tf, docLength, df, N);
            tripleList.push_back(Eigen::Triplet<float>(stemid, i, tfidf));
        }
        //        delete rel_seed_dvs[i];
    }

    Eigen::SparseMatrix<float>* rel_seeds_stem_document_matrix = new Eigen::SparseMatrix<float>(local_stem_list.size(), rel_seed_dvs.size());
    rel_seeds_stem_document_matrix->setFromTriplets(tripleList.begin(), tripleList.end());

    
    for (it=query_terms.begin() ; it != query_terms.end(); it++ )
    {
        std::string term = (*it);
        std::string stem = stem_it(term);

        // If query stem does not appear in top documents,
        // ignore it
        if(local_stem_dict.find(stem) != local_stem_dict.end())
        {
            //float prc = compute_weighted_pseudo_recall(stem);
            /* Here for recall tf*/
            //float prc = compute_pseudo_recall_tf(stem);
            //float prc = compute_pseudo_recall(stem);
            //float prc = compute_weighted_pseudo_recall_tf(stem);
            //float prc = compute_rocchio_weight(stem, 0);
            //float prc = compute_relevance_weight(stem, 0);
            float prc = compute_guided_pseudo_recall(stem, n_rel, rel_seed_dvs, rel_seeds_stem_document_matrix);
            //float prc = compute_true_recall(stem);
            //float rc = compute_true_recall(stem);

            //float idf = compute_idf(stem);
            oss<<fixed<<setprecision(16)<<prc<<" "<<term<<" ";

            float has_stem = 0;
            for(int i=0;i<rel_seed_dvs.size();i++)
            {
                for(int j=0;j<rel_seed_dvs[i]->stems().size();j++)
                {
                    if(stem.compare(rel_seed_dvs[i]->stems()[j])==0)
                    {
                        has_stem += 1.0;
                        break;
                    }
                }
            }

            float partial_true_recall = (has_stem + 1) / (rel_seed_dvs.size() + 2);
            partial_true_out<<fixed<<setprecision(16)<<partial_true_recall<<" "<<term<<" ";

            //oss<<prc<<" "<<term<<" ";

            //            oss<<prc<<" 1:"<<rc<<" 2:"<<idf<<" #"<<stem<<" "<<term<<" "<<queryid<<endl;
        }
    }

    for(int i =0;i<rel_seed_dvs.size();i++)
    {
        delete rel_seed_dvs[i];
    }

    oss<<" )"<<endl;
    oss<<"\t</text></query>"<<endl;

    partial_true_out<<" )"<<endl;
    partial_true_out<<"\t</text></query>"<<endl;
    partial_true_out.close();

    return oss.str();
}
