/*==========================================================================
 * Copyright (c) 20014 Carnegie Mellon University.  All Rights Reserved.
 *
 * Use of the Lemur Toolkit for Language Modeling and Information Retrieval
 * is subject to the terms of the software license set forth in the LICENSE
 * file included with this software (and below), and also available at
 * http://www.lemurproject.org/license.html
 *
 *==========================================================================
*/

#include "plsa.h"
#include <math.h>
#include <iostream>

PLSA::PLSA(const Eigen::SparseMatrix<float> &data, int k)
    : m_H(data)
    , K(k)
    , N(data.rows())
    , V(data.cols())
    , pi(K)
    , theta(K,V)
    , Q(N,K)
{ 
}

void PLSA::initialize()
{
	int i,j,k;
	for ( k=0; k<K; k++ )
	{
		pi[k] = 1./K;
        for(j=0;j<V;j++)
            theta(k,j)=1.0/V;

        for(i=0;i<N;i++)
            Q(i,k)=1.0/K;
	}
}

void PLSA::train(int max_iter)
{
    for(int iter = 0;iter<max_iter;iter++)
    {
        estep();
        mstep();
    }
}

void PLSA::estep()
{
	int i,j,k;
	float tmp,normal, tmp0;
	for( i=0; i<N; i++ )
	{
		for(normal=0.0, k=0; k<K; k++ )
		{
			for (tmp=0.0, j=0; j<V; j++ )
			{
                // ??
                if(m_H.coeffRef(i,j)!=0)// && pi[k]!=0)
                {
                    // Prevent underflow
                    tmp0 = m_H.coeffRef(i, j) * log(theta(k,0));
                    tmp +=  ( m_H.coeffRef(i, j) * log(theta(k,j)) - tmp0);
                }
			}
			Q(i,k) = pi[k] * exp(tmp);
			normal += Q(i,k);
		}

		for( k=0; k<K; k++ )
        {
			Q(i, k) /= normal;
        }
	}
}

void PLSA::mstep()
{
	int i,j,k;
	float tmppi,tmptheta,normal;

	for( k=0; k<K; k++ )
	{
		for( tmppi=0.0, i=0; i<N; i++ )
		{
			tmppi += Q(i,k);		
		}
		pi[k] = tmppi/N;
		for(normal=0.0, j=0; j<V; j++ )
		{
			for( tmptheta=0.0, i=0; i<N; i++ )
			{
				tmptheta += Q(i,k) * m_H.coeffRef(i, j);
			}
			theta(k,j) = tmptheta;
			normal += tmptheta;
		}

		for( j=0; j<V; j++ )
			theta(k,j) /= normal;
	}
}

Eigen::MatrixXf PLSA::getTheta()
{
    return theta;
}
