# ==========================================================================
#  Copyright (c) 20014 Carnegie Mellon University.  All Rights Reserved.
# 
#  Use of the Lemur Toolkit for Language Modeling and Information Retrieval
#  is subject to the terms of the software license set forth in the LICENSE
#  file included with this software (and below), and also available at
#  http://www.lemurproject.org/license.html
# 
# ==========================================================================

import os
import sys
import math
import copy
import subprocess
import cPickle
import random
from datetime import datetime
from collections import defaultdict

INDRIRUNQUERY = '/bos/usr0/gzheng/.local/bin/IndriRunQuery'
INDRIGETSCORE = '/bos/usr0/gzheng/TermRecall/wsd/src/IndriGetScore'
BM25GETSCORE = '/bos/usr0/gzheng/TermRecall/wsd/src/BM25GetScore'

MU = 2500
EPSILON = 0.0000001

def get_test_tid_cikm(tid_str):
    if tid_str == 'robust04':
        return 7
    elif tid_str == 'wt10g':
        return 9
    elif tid_str == 'gov2':
        return 13
    elif tid_str == 'clueweb09b':
        return 21

def lm_score(tf, cf, C, D, mu):
    logrand = ( tf + float(mu) * cf / C ) / ( D + mu)
    if logrand == 0:
        score = sys.float_info.min
    else:
        score = math.log( ( tf + float(mu) * cf / C ) / ( D + mu))
    return score

def bm25_score(tf, df, N, dl, avgdl):
    k1 = 1.2
    b = 0.75
    score = math.log(1.0*(N-df + 0.5)/(df+0.5)) * tf * (k1+1) / (tf+k1*(1-b+b*dl/avgdl))
    return score

def get_matching_scores_bm25(index, res_file, qid, concept):
    scores = defaultdict(dict)
    cmd = '%s %s %s %d %s' % (BM25GETSCORE, index, res_file, qid, concept)
    output = subprocess.check_output(cmd.split(' ', 4))
    lines = output.strip().split('\n')
    
    for line in lines:
        parts = line.split()
        docno = parts[0].strip()
        score = float(parts[1])
        scores[docno] = score

    return scores

def get_matching_scores(index, res_file, qid, concept):
    scores = defaultdict(dict)
    cmd = '%s %s %s %d %s' % (INDRIGETSCORE, index, res_file, qid, concept)
    output = subprocess.check_output(cmd.split(' ', 4))
    lines = output.strip().split('\n')

    for line in lines:
        parts = line.split()
        docno = parts[0].strip()
        score = float(parts[1])
        scores[docno] = score

    return scores

def read_all_features(wsd_file_name):
    all_concepts = defaultdict(set)
    all_features = defaultdict(dict)

    feat_dim = None
    with open(wsd_file_name, 'r') as wsd_file:
        for line in wsd_file:
            qid = int(line.strip().split('\t')[-1])
            concept = line.strip().split('\t')[-2].split('#', 1)[1].strip()

            all_concepts[qid].add(concept)
            feats = line.strip().split('#', 1)[0].strip().split(' ')
            feats_value = []
            feat_dim = 0;
            for feat in feats:
                if feat != '':
                    feats_value.append(float(feat))
                    feat_dim += 1

            unpacked_feats_value = feats_value

            # according to WSD, features for different concepts are in different dimensions
            # START unpacking
            if concept.startswith('#1('): # bigram
                unpacked_feats_value =  [0 for i in range(feat_dim - 1)] + feats_value[:feat_dim-1] + [0 for i in range(feat_dim-1)] + [1]
            elif concept.startswith('#uw8('): #uw8
                unpacked_feats_value =  [0 for i in range(2 * feat_dim-2)] + feats_value[:feat_dim-1] + [1]
            else: # term
                unpacked_feats_value =  feats_value[:feat_dim-1] + [0 for i in range(2 * feat_dim-2)] + [1]
            feat_dim = 3 * feat_dim - 2
            # END unpacking

            (all_features[qid])[concept] = unpacked_feats_value

    return (all_concepts, all_features, feat_dim)

# normalize features for both train and testing
def normalize_all_features(train_features, test_features, feat_dim):
    for feat_id in range(0, feat_dim):
        mean = 0.0
        n_instance = 0

        for qid in train_features:
            for concept in train_features[qid]:
                mean += ((train_features[qid])[concept])[feat_id]
                n_instance += 1

        for qid in test_features:
            for concept in test_features[qid]:
                mean += ((test_features[qid])[concept])[feat_id]
                n_instance += 1

        mean /= n_instance
        std = 0.0

        for qid in train_features:
            for concept in train_features[qid]:
                x = ((train_features[qid])[concept])[feat_id] - mean
                std += x*x

        for qid in test_features:
            for concept in test_features[qid]:
                x = ((test_features[qid])[concept])[feat_id] - mean
                std += x*x

        std = math.sqrt( std / (n_instance -1))
        
        # normalize
        if std > 0:
            for qid in train_features:
                for concept in train_features[qid]:
                    raw = ((train_features[qid])[concept])[feat_id]
                    ((train_features[qid])[concept])[feat_id] = (raw - mean) / std

            for qid in test_features:
                for concept in test_features[qid]:
                    raw = ((test_features[qid])[concept])[feat_id]
                    ((test_features[qid])[concept])[feat_id] = (raw - mean) / std
        
    return (train_features, test_features)
    
    
def compute_all_matching_scores(index, res_file, concepts, rmmodel):
    pickle_file_name = '../wsd_pickles/%s.%s.pk'%(res_file.split('/')[-1], rmmodel)

    if os.path.exists(pickle_file_name) and os.path.getsize(pickle_file_name) > 0:
        with open(pickle_file_name, 'rb') as pk_in:
            all_matching_scores = cPickle.load(pk_in)
            return all_matching_scores
        
    all_matching_scores = dict()
    for qid in concepts:
        print '=== %d ===' % qid
        query_matching_scores = dict()
        for concept in concepts[qid]:
            if rmmodel == 'LM':
                query_matching_scores[concept] = get_matching_scores(index, res_file, qid, concept)
            else: #BM25
                query_matching_scores[concept] = get_matching_scores_bm25(index, res_file, qid, concept)

        all_matching_scores[qid] = query_matching_scores

    with open(pickle_file_name, 'wb') as pk_out:
        cPickle.dump(all_matching_scores, pk_out)
    return all_matching_scores
            

#get_feature('hello')
#get_feature('nice weather')

def compute_map(res, qrel):
    # res is a defaultdict of list ( ordered results)
    # qrel is a defaultdict of set ( set of relevant docs)
    mean_ap = 0.0

    for qid in res:
        ap = 0.0
        if not qid in qrel:
            #print 'ERROR: Query %s not in relevance judgements!' % qid
            continue

        num_rel = 0
        for idx, docno in enumerate(res[qid]):
            rank = idx + 1

            if rank > 1000:
                break

            if docno in qrel[qid]: # relevant
                num_rel += 1
                ap += (float(num_rel) / rank)

        ap /= len(qrel[qid])
        mean_ap += ap

    mean_ap /= len(qrel)
#    print 'Num of queries: %d MAP: %f' % (len(res), mean_ap)

    return mean_ap


# Res file is like:
# 151 Q0 AP881005-0001 1 -5.63196 indri

def read_res(res_file):
    res = defaultdict(list)
    with open(res_file, 'r') as res_in:
        for line in res_in:
            parts = line.split(' ')
            qid = int(parts[0])
            docno = parts[2].strip()
            res[qid].append(docno)

    return res
    
# Qrel file is like:
# 151 0 AP880213-0002 1

def read_qrel(qrel_file):
    qrel = defaultdict(set)
    with open(qrel_file, 'r') as qrel_in:
        for line in qrel_in:
            parts = line.split()
            qid = int(parts[0])
            docno = parts[2].strip()
            rel = int(parts[3].strip())
            if rel > 0:  #relevant judgement
                qrel[qid].add(docno)

    return qrel

# training in wsd requires to append non-retrieved relevant docs to res
def append_qrel_to_res(res, qrel, res_path):

    local_qrel = copy.deepcopy(qrel)
    new_res = defaultdict(list)
    new_res_filename =  '%s.append' % res_path

    new_res_w = open(new_res_filename, 'w')
    
    for qid in res:
        for docno in res[qid]:
            new_res[qid].append(docno)
            new_res_w.write('%d Q0 %s -1 indri\n' % (qid, docno))
            if docno in local_qrel[qid]:
                (local_qrel[qid]).remove(docno)

        for docno in local_qrel[qid]:
            new_res[qid].append(docno)
            new_res_w.write('%d Q0 %s -1 indri\n' % (qid, docno))

    new_res_w.close()
    return (new_res, new_res_filename)


# ../../queries/desc_parts_stopped/322.part
def read_concepts_per_query(query_part_dir, qid):
    concept_per_query = []
    with open('%s/%d.part' % (query_part_dir, qid), 'r') as concepts_in:
        for line in concepts_in:
            concept_per_query.append(line.strip())

    return concept_per_query

'''
res = defaultdict(list)
qrel = defaultdict(set)
#res[1]=[1,2,3,4,5]
#res[2] =[1,2,3]
#qrel[1] = [6]
#qrel[2] = [3,2,1]
res = read_res('../../res/t3.okw.mu900.res')
qrel = read_qrel('../../qrels/qrels.t3.txt')
compute_map(res,qrel)
'''

def inner_product(ws, feature):
    val = 0.0
    for idx, w in enumerate(ws):
        val += w * feature[idx]

    return val

def rank_res(input_ws, all_features, all_matching_scores):
    #tick = datetime.now()

    # renormalize ws first
    ws = normL1(input_ws)

    new_res= defaultdict(list)
    for qid in all_matching_scores:
        scores = defaultdict(float)
        tuples = []

        num_term = 0
        num_bigram = 0
        num_uw = 0
        for concept in all_matching_scores[qid]:
            if concept.startswith('#1('):
                num_bigram += 1
            elif concept.startswith('#uw8('):
                num_uw += 1
            else:
                num_term += 1
        
        for concept in all_matching_scores[qid]:
            feature = (all_features[qid])[concept]
            weight = inner_product(ws, feature)
            if concept.startswith('#1('):
                weight += 0.1
                weight /= num_bigram
            elif concept.startswith('#uw8('):
                weight += 0.1
                weight /= num_uw
            else:
                weight += 0.8
                weight /= num_term

#            if weight < 0:
#                print 'NEGATIVE WEIGHTS!!!!!!!!'
#                weight = EPSILON
            
            for docno in (all_matching_scores[qid])[concept]:
                score = ((all_matching_scores[qid])[concept])[docno]  
                scores[docno] += weight * score

        for docno in scores.keys():
            tuples.append((scores[docno], docno))

        tuples.sort(reverse=True)
        
        for x in tuples:
            new_res[qid].append(x[1])


    #tock = datetime.now()
    #diff = tock - tick
    #print 'Rank res time: ', diff.microseconds

    return new_res

def normL1(w):
    norm = math.fsum(abs(s) for s in w)
    new_w = []
    if norm > 0:
        for s in w:
            new_w.append(s/norm)

    return new_w

# coord_ascent from Michael
def coord_ascent_michael(feat_dim, qrel, all_features, all_matching_scores):
    # some constants
    n_Restart = 5
    minStep = 0.02
    stepScale = 2.0
    maxStepRatio = 0.5
    
#    random.seed(1234871234) # for reproducible results
    r = 0
    optimized = True

    weight = []
    for i in range(0, feat_dim):
        weight.append(0)

    startRes = rank_res(weight, all_features, all_matching_scores)
    startScore = compute_map(startRes, qrel)
    
    # local best (within the current restart cycle)
    bestModelScore = startScore
    bestModel = weight

    print 'Initial parameter weights:', bestModel, ' MAP: ', bestModelScore, 'Starting optimization...'
    while (optimized and r < n_Restart):
        optimized = False
        
        print 'Restart #', (r+1), '/', n_Restart, '...'
        r += 1
        
        fids = range(0, len(weight))
        random.shuffle(fids)
        
        for i in range(0, len(fids)):
            current_feature = fids[i]
            print 'Trying feature', current_feature
            origWeight = weight[current_feature]

            # Take a step to the right
            step = minStep
            if origWeight !=0 and step > maxStepRatio * abs(origWeight):
                step = maxStepRatio * abs(origWeight)

            rightBest = bestModelScore
            rightStep = 0
            change = True
            while (change):
                weight[current_feature] += step
                currentRes = rank_res(weight, all_features, all_matching_scores)
                currentScore = compute_map(currentRes, qrel)
                if currentScore > rightBest:
                    rightBest = currentScore
                    rightStep += step
                    step *= stepScale
                else:
                    change = False

            weight[current_feature] = origWeight # revert change

            # Take a step to the left
            step = minStep
            if origWeight !=0 and step > maxStepRatio * abs(origWeight):
                step = maxStepRatio * abs(origWeight)

            leftBest = bestModelScore
            leftStep = 0
            change = True
            while (change):
                weight[current_feature] -= step
                currentRes = rank_res(weight, all_features, all_matching_scores)
                currentScore = compute_map(currentRes, qrel)
                if currentScore > leftBest:
                    leftBest = currentScore
                    leftStep += step
                    step *= stepScale
                else:
                    change = False

            weight[current_feature] = origWeight # revert change

            # Pick a direction to go
            if rightBest > leftBest:
                optimized = True
                weight[current_feature] += rightStep
                bestModelScore = rightBest

                if len(weight) > 1:
                    weight = normL1(weight)
                bestModel = weight
                print 'Finished optimizing coordinate (%d). ++%f. Metric: %f' %(current_feature, rightStep, bestModelScore)
            elif leftBest > rightBest:
                optimized = True
                weight[current_feature] -= leftStep
                bestModelScore = leftBest

                if len(weight) > 1:
                    weight = normL1(weight)
                bestModel = weight
                print 'Finished optimizing coordinate (%d). --%f. Metric: %f' %(current_feature, leftStep, bestModelScore)
            else:
                print 'Finished optimizing coordinate (%d). No Change.' %(current_feature)

    print 'Coord ascent done.'
    return bestModel

# similar implementation to vdang's Java code
def coord_ascent(feat_dim, qrel, all_features, all_matching_scores):
    # some constants
    n_Restart = 5
    n_MaxIteration = 25
    stepBase = 0.05
    stepScale = 2.0
    tolerance = 0.001
    regularized = False
    slack = 0.001
    
    bestModel = None
    bestModelScore = 0.0

    signs = [1, -1] # two directions
    
    #random.seed(1234871234) # for reproducible results
    
    weight = []
    for i in range(0, feat_dim):
        weight.append(0)

    for r in range(0, n_Restart):
        print 'Random restart #', (r+1), '/', n_Restart, '...'
        consecutive_fails = 0
        
        if r==0: # start from uniform weights 
            for i in range(0, feat_dim):
                weight[i] = 1.0 / feat_dim
        else:
            sum_w = 0
            for i in range(0, feat_dim):
                weight[i] = random.random()
                sum_w += weight[i]

            for i in range(0, feat_dim):
                weight[i] /= sum_w

        for i in range(0, feat_dim):
            weight[i] = 0
    
        current_feature = -1;
        startRes = rank_res(weight, all_features, all_matching_scores)
        startScore = compute_map(startRes, qrel)
        
        # local best (within the current restart cycle)
        bestLocalScore = startScore
        bestLocalWeight = weight

#        while (len(weight) > 1 and consecutive_fails < len(weight) -1 ) or (len(weight)==1 and consecutive_fails ==0): # why?
        while consecutive_fails < len(weight): # do as long as not all features fail
            # shuffle  feature index
            fids = range(0, len(weight))
            random.shuffle(fids)

            for i in range(0, len(fids)):
                current_feature = fids[i]
                print 'Trying feature', current_feature, 'Consecutive fails = ', consecutive_fails
                
                origWeight = weight[current_feature]
                totalStep = 0
                bestTotalStep = 0
                succeeds = False

                for s in range(0, len(signs)):
                    step = 0.001 * signs[s]
                    if origWeight != 0 and abs(step) > 0.5 * abs(origWeight):
                        step = stepBase * abs(origWeight)

                    totalStep = step
                    
                    for j in range(0, n_MaxIteration):
                        w = origWeight + totalStep
                        weight[fids[i]] = w
                        currentRes = rank_res(weight, all_features, all_matching_scores)
                        currentScore = compute_map(currentRes, qrel)
                        
                        if(currentScore > bestLocalScore):
                            bestLocalScore = currentScore
                            bestTotalStep = totalStep
                            succeeds = True
                            print 'New update:', bestLocalScore, normL1(weight)
                            
                        # shrink step size and move one step forward    
                        step *= stepScale
                        totalStep += step

                    if succeeds:
                        weight[fids[i]] = origWeight + bestTotalStep

                        weight = normL1(weight)
                        
                        consecutive_fails = 0
                        bestLocalWeight = weight
                        break # this direction helps
                    else: # recover original weight for this feature after this direction
                        weight[fids[i]] = origWeight

                if not succeeds: # failed all directions for this feature
                    consecutive_fails += 1

            # not much progress
            if bestLocalScore - startScore < tolerance:
                break

        if bestModel == None or bestLocalScore > bestModelScore:
            bestModelScore = bestLocalScore
            bestModel = bestLocalWeight

    print 'Coord ascent done.'
    return bestModel

def construct_and_search(ws, features, trecno, rmmodel):
    query_file = '../tmp_wsd/t%d.desc.wsd.%s.q' % (trecno, rmmodel)
    res_file = '../res_wsd/t%d.desc.wsd.%s.res' % (trecno, rmmodel)
    
    with open(query_file, 'w') as f:
        f.write('<parameters>\n')
        
        for qid in features:
            f.write('\t<query><type>indri</type><number>%d</number><text>\n' % qid)
            f.write('#weight(\n')


            num_term = 0
            num_bigram = 0
            num_uw = 0
            for concept in features[qid]:
                if concept.startswith('#1('):
                    num_bigram += 1
                elif concept.startswith('#uw8('):
                    num_uw += 1
                else:
                    num_term += 1

            for concept in features[qid]:
                feature = (features[qid])[concept];
                computed_weight = inner_product(ws, feature)
                if concept.startswith('#1('):
                    computed_weight += 0.1
                    computed_weight /= num_bigram
                elif concept.startswith('#uw8('):
                    computed_weight += 0.1
                    computed_weight /= num_uw
                else:
                    computed_weight += 0.8
                    computed_weight /= num_term

#                if computed_weight < 0:
#                    computed_weight = EPSILON
                    
                f.write('         %.16f %s\n' % (computed_weight, concept))

            f.write(')\n')
            f.write('\t</text></query>\n')

        f.write('</parameters>\n')

    # search
    rulestr = None
    if rmmodel == 'LM':
        rulestr = '-rule=method:d,mu:2500'
    else:
        rulestr = '-rule=method:okapi'
    
    search_cmd = '%s -count=1000 %s -index=/bos/tmp17/gzheng/trec/indext%d-v5.3/ -trecFormat=t %s > %s' % (INDRIRUNQUERY, query_file, trecno, rulestr, res_file)
    os.system(search_cmd)

    # eval
    eval_output = subprocess.check_output(('./trec_eval ../qrels/qrels.t%d.txt %s' %(trecno, res_file)).split(' '))
    eval_map = float(eval_output.strip().split('\n')[5].split('\t')[-1])
    eval_p10 = float(eval_output.strip().split('\n')[22].split('\t')[-1])
    return (eval_map, eval_p10)

def write_res_to_tofile(res, filename):
    with open(filename, 'w') as w:
        for qid in res:
            for idx, docno in enumerate(res[qid]):
                score = 1000 - idx
                rank = idx + 1
                if rank > 1000:
                    break
                w.write('%d Q0 %s %d %d wsd\n' % (qid, docno, rank, score))
            
if __name__ == "__main__":
    if len(sys.argv)!=5:
        print 'Usage: %s <trec_str> <train_wsd_file> <test_wsd_file> <LM|BM25>' % sys.argv[0]
        exit(0)
        
        
    rmmodel = sys.argv[4]
    if rmmodel != 'LM' and rmmodel != 'BM25':
        print 'Usage: %s <train_trecid> <test_trecid> <train_wsd_file> <test_wsd_file> <LM|BM25>' % sys.argv[0]
        exit(0)
    
    trec_str = sys.argv[1]
    train_tid = get_test_tid_cikm(trec_str)
    test_tid = train_tid
    
    train_index = '/bos/tmp17/gzheng/trec/indext%d-v5.3/' % train_tid
    test_index = '/bos/tmp17/gzheng/trec/indext%d-v5.3/' % test_tid


    if rmmodel == 'LM':
        train_res_file = '/bos/usr0/gzheng/TermRecall/src/%s.sd.res' % trec_str
        test_res_file  = '/bos/usr0/gzheng/TermRecall/src/%s.sd.res' % trec_str
    else: # BM25
        train_res_file = '/bos/usr0/gzheng/TermRecall/src/%s.sd.bm25.res' % trec_str
        test_res_file  = '/bos/usr0/gzheng/TermRecall/src/%s.sd.bm25.res' % trec_str

    train_res = read_res(train_res_file)
    test_res = read_res(test_res_file)
    
    train_qrel = read_qrel('../qrels/qrels.%s.txt' % trec_str)
    test_qrel = read_qrel('../qrels/qrels.%s.txt' % trec_str)

    print 'Initial train res MAP: %.4f %s'% (compute_map(train_res, train_qrel), rmmodel)
    print 'Initial test res MAP: %.4f %s'% (compute_map(test_res, test_qrel), rmmodel)

    # append non-retrieved rel docs to res and note res_file path will be changed!!!
    (train_res, train_res_file) = append_qrel_to_res(train_res, train_qrel, train_res_file)
    (test_res, test_res_file) = append_qrel_to_res(test_res, test_qrel, test_res_file)    

    train_wsd_file  = sys.argv[2]
    test_wsd_file = sys.argv[3]

    # read raw features
    (train_concepts, train_all_features, feat_dim) = read_all_features(train_wsd_file)
    (test_concepts, test_all_features, feat_dim) = read_all_features(test_wsd_file)

    # normalize features
    (train_all_features, test_all_features) = normalize_all_features(train_all_features, test_all_features, feat_dim)


    train_all_matching_scores = compute_all_matching_scores(train_index, train_res_file, train_concepts, rmmodel)
    ws = coord_ascent(feat_dim, train_qrel, train_all_features, train_all_matching_scores)
#    ws = coord_ascent_michael(feat_dim, train_qrel, train_all_features, train_all_matching_scores)
    #ws = [0.19266666666666665, -0.13033333333333333, -0.12133333333333333, 0.009666666666666667, -0.18833333333333335, 0.08366666666666667, 0.6706666666666666, 0.13466666666666666, 0.0015429166666666634, 0.014666666666666661, 0.12666666666666665, 0.06566666666666665, 0.07966666666666666, 0.06466666666666668, 0.41866666666666663]

    train_best_res = rank_res(ws, train_all_features, train_all_matching_scores)
    train_best_map = compute_map(train_best_res, train_qrel)
    print 'Best MAP on train: %.4f %s' % (train_best_map, rmmodel)

    test_all_matching_scores = compute_all_matching_scores(test_index, test_res_file, test_concepts, rmmodel)
    test_predicted_res = rank_res(ws, test_all_features, test_all_matching_scores)
    test_predicted_map = compute_map(test_predicted_res, test_qrel)
    print 'Predicted MAP on test: %.4f %s' % (test_predicted_map, rmmodel)

    write_res_to_tofile(test_predicted_res, '%s.rerank.res' % test_wsd_file)
    eval_output = subprocess.check_output(('./trec_eval ../qrels/qrels.%s.txt %s.rerank.res' %(trec_str, test_wsd_file)).split(' '))
    test_map = float(eval_output.strip().split('\n')[5].split('\t')[-1])
    test_p10 = float(eval_output.strip().split('\n')[22].split('\t')[-1])

    # construct query and do real retrieval!
    # (test_map, test_p10) = construct_and_search(ws, test_all_features, test_tid, rmmodel)
    print 'FINAL RESULT: %s %s\t%.4f\t%.4f %s' % (train_wsd_file, test_wsd_file, test_map, test_p10, rmmodel)
                        
